/*
 * Name:    mxcaMcaRecord.h
 *
 * Purpose: This header file defines the mxcaMcaRecord class.
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 2012 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

extern "C" {
#include "mx_mca.h"
}

class mxcaMcaAcquiringPV;
class mxcaMcaEraseAllPV;
class mxcaMcaEraseStartPV;
class mxcaMcaPresetLivePV;
class mxcaMcaPresetRealPV;
class mxcaMcaStartAllPV;
class mxcaMcaStopAllPV;
class mxcaMcaChValPV;

/*------------------------------------------------------------------------*/

#define MXCA_MCA_MAX_MCAS	64

typedef struct {
	MX_RECORD *mx_record;
	int channel_number;

	mxcaMcaChValPV *val_pv;
} mxcaMcaChannelStruct;

/*------------------------------------------------------------------------*/

class mxcaMcaRecord : public mxcaRecord {
    friend class mxcaMcaPresetLivePV;
    friend class mxcaMcaPresetRealPV;

    public:
    	mxcaMcaRecord( mxcaServer *epics_server,
			const char *epics_record_name );

	mx_status_type add_mca( MX_RECORD *mx_mca_record );

	mxcaMcaChannelStruct *get_mca_channel( int channel_number );

	mx_status_type setup_global_pvs( int master_channel_number );

    private:
	mxcaMcaAcquiringPV	*acquiring_pv;

	mxcaMcaStartAllPV	*start_all_pv;
	mxcaMcaEraseAllPV	*erase_all_pv;
	mxcaMcaEraseStartPV	*erase_start_pv;
	mxcaMcaStopAllPV	*stop_all_pv;

	mxcaMcaPresetLivePV	*preset_live_pv;
	mxcaMcaPresetRealPV	*preset_real_pv;

	MX_RECORD *mx_timer_record;

	int num_mca_channels;
	mxcaMcaChannelStruct mca_array[MXCA_MCA_MAX_MCAS];
};

/*------------------------------------------------------------------------*/

class mxcaMcaAcquiringPV : public mxcaPV {
    public:
	mxcaMcaAcquiringPV( mxcaRecord *epics_record );

    private:
};

/*------------------------------------------------------------------------*/

class mxcaMcaEraseAllPV : public mxcaPV {
    public:
	mxcaMcaEraseAllPV( mxcaRecord *epics_record );

    private:
};

/*------------------------------------------------------------------------*/

class mxcaMcaEraseStartPV : public mxcaPV {
    public:
	mxcaMcaEraseStartPV( mxcaRecord *epics_record );

    private:
};

/*------------------------------------------------------------------------*/

class mxcaMcaPresetLivePV : public mxcaPV {
    public:
	mxcaMcaPresetLivePV( mxcaRecord *epics_record );

    private:
};

/*------------------------------------------------------------------------*/

class mxcaMcaPresetRealPV : public mxcaPV {
    public:
	mxcaMcaPresetRealPV( mxcaRecord *epics_record );

    private:
};

/*------------------------------------------------------------------------*/

class mxcaMcaStartAllPV : public mxcaPV {
    public:
	mxcaMcaStartAllPV( mxcaRecord *epics_record );

    private:
};

/*------------------------------------------------------------------------*/

class mxcaMcaStopAllPV : public mxcaPV {
    public:
	mxcaMcaStopAllPV( mxcaRecord *epics_record );

    private:
};

/*========================================================================*/

class mxcaMcaChValPV : public mxcaPV {
    public:
	mxcaMcaChValPV( mxcaRecord *epics_record,
			int channel_number,
			const char *epics_pv_name );

	caStatus read( void );

    private:
};

