/*
 * Name:    mxcaMotorRecord.cpp
 *
 * Purpose: This file defines the mxcaMotorRecord class.
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 2004, 2008-2009, 2011-2014 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXCA_MOTOR_DEBUG				FALSE

#define MXCA_MOTOR_DEBUG_USER_AND_DIAL			FALSE

#define MXCA_MOTOR_DEBUG_MSTA				FALSE

#define MXCA_MOTOR_DEBUG_DIAL_POST_VALUE_CHANGED	FALSE

#include <iostream>
#include <fstream>
#include <cfloat>

#include "mxcaServer.h"
#include "mxcaMotorRecord.h"

extern "C" {
#include "mx_driver.h"
}

mxcaMotorRecord::mxcaMotorRecord( mxcaServer *epics_server,
				const char *epics_record_name,
				MX_RECORD *mx_record ) :
					mxcaRecord( epics_server,
						epics_record_name,
						mx_record )
{
#if MXCA_MOTOR_DEBUG
	static const char fname[] = "mxcaMotorRecord::mxcaMotorRecord()";

	MX_DEBUG(-2,("%s invoked for epics motor record '%s'",
		fname, epics_record_name));
#endif

	short short_value;
	float float_value;
	aitEnum16 enum_value;

	/*---*/

	this->set_pv = new mxcaEnumPV( this, ".SET" );

	enum_value = 0;
	this->set_pv->configure_enum_pv( enum_value, 2, "Use", "Set" );

	this->set_value = &(this->set_pv->enum_value);

	/*---*/

	this->foff_pv = new mxcaEnumPV( this, ".FOFF" );

	enum_value = 0;
	this->foff_pv->configure_enum_pv( enum_value, 2, "Variable", "Frozen" );

	this->foff_value = &(this->foff_pv->enum_value);

	/*---*/

	this->dir_pv = new mxcaMotorDirPV( this );

	this->dir_enum_value = &(this->dir_pv->enum_value);

	this->off_pv = new mxcaMotorOffPV( this );

	this->off_value = (double *) this->off_pv->get_mx_value_pointer();

	this->eres_pv = new mxcaPV( this, ".ERES", "scale" );

	this->mres_pv = new mxcaPV( this, ".MRES", "scale" );

	/*---*/

	this->dhlm_pv = new mxcaMotorDialPV( this,
					".DHLM", "positive_limit", FALSE );

	this->dllm_pv = new mxcaMotorDialPV( this,
					".DLLM", "negative_limit", FALSE );

	this->drbv_pv = new mxcaMotorDialPV( this, ".DRBV", "position", TRUE );

	this->hlm_pv = new mxcaMotorUserPV( this, ".HLM", dhlm_pv );

	this->llm_pv = new mxcaMotorUserPV( this, ".LLM", dllm_pv );

	this->rbv_pv = new mxcaMotorUserPV( this, ".RBV", drbv_pv );

	/*---*/

	this->dval_pv = new mxcaMotorDvalPV( this );

	this->dval_value = (double *) this->dval_pv->mx_value;

	this->val_pv = new mxcaMotorValPV( this );

	this->rlv_pv = new mxcaMotorRlvPV( this );

	/* The mxcaMotorMstaPV constructor also creates the
	 * ATHM, DMOV and MOVN PVs.
	 */

	this->msta_pv = new mxcaMotorMstaPV( this );

	/*---*/

	this->accl_pv = new mxcaPV( this, ".ACCL", "acceleration_time" );

	this->acof_pv = new mxcaPV( this, ".ACOF",
					"acceleration_feedforward_gain" );

	this->bacc_pv = new mxcaPV( this, ".BACC", "acceleration_time" );

	this->bdst_pv = new mxcaPV( this, ".BDST", "backlash_correction" );

	this->card_pv = new mxcaDummyPV( this, ".CARD" );
	short_value = -1;
	this->card_pv->configure_internal_pv( aitEnumInt16,
						0, NULL, &short_value );

	this->dcof_pv = new mxcaPV( this, ".DCOF", "derivative_gain" );

	this->diff_pv = new mxcaMotorDiffPV( this );

	this->dtyp_pv = new mxcaPV( this, ".DTYP" );
	this->dtyp_pv->configure_internal_pv( aitEnumString,
						0, NULL, "MxMotor" );

	this->egu_pv = new mxcaPV( this, ".EGU", "units" );

	this->hls_pv = new mxcaPV( this, ".HLS", "positive_limit_hit" );

	this->homf_pv = new mxcaMotorHomePV( this, ".HOMF", 1 );

	this->homr_pv = new mxcaMotorHomePV( this, ".HOMR", -1 );

	this->icof_pv = new mxcaPV( this, ".ICOF", "integral_gain" );

	this->jogf_pv = new mxcaMotorJogPV( this, ".JOGF", 1 );

	this->jogr_pv = new mxcaMotorJogPV( this, ".JOGR", -1 );

	this->lls_pv = new mxcaPV( this, ".LLS", "negative_limit_hit" );

	this->lvio_pv = new mxcaPV( this, ".LVIO" );

	this->lvio_value = new short;
	*(this->lvio_value) = 0;
	this->lvio_pv->configure_internal_pv( aitEnumInt16,
						0, NULL, this->lvio_value );

	this->mdel_pv = new mxcaPV( this, ".MDEL", "value_change_threshold" );

	this->miss_pv = new mxcaPV( this, ".MISS" );

	short_value = 0;
	this->miss_pv->configure_internal_pv( aitEnumInt16,
						0, NULL, &short_value );

	this->pcof_pv = new mxcaPV( this, ".PCOF", "proportional_gain" );

	/* MX currently does not do retries, but OpenGDA expects
	 * the RDBD PV to be present. (WML 2012-01-20).
	 */

	this->rdbd_pv = new mxcaDummyPV( this, ".RDBD" );
	short_value = 0;
	this->rdbd_pv->configure_internal_pv( aitEnumInt16,
						0, NULL, &short_value );

	this->rep_pv = new mxcaMotorRxpPV( this, ".REP" );

	this->rep_pv = new mxcaMotorRxpPV( this, ".RMP" );

	this->rrbv_pv = new mxcaMotorRrbvPV( this );

	this->rtry_pv = new mxcaDummyPV( this, ".RTRY" );
	short_value = 0;
	this->rdbd_pv->configure_internal_pv( aitEnumInt16,
						0, NULL, &short_value );

	this->rval_pv = new mxcaPV( this, ".RVAL", "raw_destination" );

	/* FIXME - STAT is not documented.  I do not know what it
	 * is supposed to do, so I have hard coded it to be 0.
	 */

	this->stat_pv = new mxcaDummyPV( this, ".STAT" );
	short_value = 0;
	this->stat_pv->configure_internal_pv( aitEnumInt16,
						0, NULL, &short_value );

	this->stop_pv = new mxcaPV( this, ".STOP", "soft_abort" );

	this->spmg_pv = new mxcaMotorSpmgPV( this );

	this->tdir_pv = new mxcaPV( this, ".TDIR" );

	this->tdir_value = new short;
	*(this->tdir_value) = 0;
	this->tdir_pv->configure_internal_pv( aitEnumInt16,
						0, NULL, this->tdir_value );

	/*---*/

	this->twf_pv = new mxcaMotorTweakPV( this, ".TWF", 1 );

	this->twr_pv = new mxcaMotorTweakPV( this, ".TWR", -1 );

	this->twv_pv = new mxcaPV( this, ".TWV" );

	this->twv_value = new double;
	*(this->twv_value) = 0.0;
	this->twv_pv->set_mx_value_pointer( this->twv_value );
	this->twv_pv->configure_internal_pv( aitEnumFloat64,
						0, NULL, this->twv_value );
	/*---*/

	this->ueip_pv = new mxcaEnumPV( this, ".UEIP" );

	enum_value = 0;
	this->ueip_pv->configure_enum_pv( enum_value, 2, "No", "Yes" );

	this->vbas_pv = new mxcaPV( this, ".VBAS", "base_speed" );

	this->vcof_pv = new mxcaPV( this, ".VCOF", "velocity_feedforward_gain");

	this->velo_pv = new mxcaPV( this, ".VELO", "speed" );

	this->bvel_pv = new mxcaPV( this, ".BVEL", "speed" );

	/* This record attempts to be compatible with version 6.3 of
	 * the EPICS motor record.
	 */

	this->vers_pv = new mxcaPV( this, ".VERS" );

	float_value = (float) 6.3;
	this->vers_pv->configure_internal_pv( aitEnumFloat32,
						0, NULL, &float_value );

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("%s complete for epics motor record '%s'",
		fname, epics_record_name));
#endif
}

void
mxcaMotorRecord::set_dval_to_drbv( void )
{
	MX_MOTOR *motor;

	motor = (MX_MOTOR *) this->mx_record->record_class_struct;

	/* Get the current motor position. */

	this->drbv_pv->read();

	this->drbv_pv->post_value_changed();

	switch( motor->subclass ) {
	case MXC_MTR_STEPPER:

#if MXCA_MOTOR_DEBUG
		MX_DEBUG(-2,("SPMG(0): raw position = %ld",
					motor->raw_position.stepper));
#endif

		motor->raw_destination.stepper = motor->raw_position.stepper;

		motor->destination = motor->offset
				+ motor->scale * motor->raw_destination.stepper;
		break;
	case MXC_MTR_ANALOG:

#if MXCA_MOTOR_DEBUG
		MX_DEBUG(-2,("SPMG(0): raw position = %g",
					motor->raw_position.analog));
#endif

		motor->raw_destination.analog = motor->raw_position.analog;

		motor->destination = motor->offset
				+ motor->scale * motor->raw_destination.analog;
		break;
	}

	this->dval_pv->post_value_changed();

	return;
}

/*-----------------------------------------------------------------------*/

mxcaMotorDvalPV::mxcaMotorDvalPV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, ".DVAL" )
{
#if MXCA_MOTOR_DEBUG
	static const char fname[] = "mxcaMotorDvalPV::mxcaMotorDvalPV()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	MX_RECORD *mx_record;
	mx_status_type mx_status;

	mx_status = mxcaPV::configure_mx_field_pv( "destination" );

	if ( mx_status.code != MXE_SUCCESS )
		return;

	/* Find the 'set_position' MX_RECORD_FIELD in the current MX record. */

	mx_record = this->epicsRecord->mx_record;

	mx_status = mx_find_record_field( mx_record, "set_position",
						&(this->set_position_rf) );

	if ( mx_status.code != MXE_SUCCESS )
		return;

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("%s: PV '%s', this->set_position_rf = %p",
		fname, this->PVname, this->set_position_rf));

	MX_DEBUG(-2,("%s: PV '%s', this->set_position_rf = '%s'",
		fname, this->PVname, this->set_position_rf->name));
#endif

	/*---*/

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return;
}

caStatus
mxcaMotorDvalPV::read( void )
{
#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
	static const char fname[] = "mxcaMotorDvalPV::read()";
#endif

	mxcaMotorRecord *mr;
	double dval, off, val;
	int dir;
	caStatus epics_status;

	mr = (mxcaMotorRecord *) this->epicsRecord;

	/* If the value of the DVAL pv has not yet been initialized, then
	 * copy its value from the DRBV pv.
	 */

#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
	MX_DEBUG(-2,("%s: PV '%s', this->get_pv_value_initialized() = %d",
		fname, this->PVname, (int) this->get_pv_value_initialized() ));
#endif

	if ( this->get_pv_value_initialized() == FALSE ) {
		mr->drbv_pv->read();

		*(mr->dval_value) = *(mr->drbv_pv->dial_value);

		this->set_pv_value_initialized( TRUE );
		mr->val_pv->set_pv_value_initialized( TRUE );
	} else {
		epics_status = mxcaPV::read();

		/* Shut up GCC 4.6 unused but set warning. */
		epics_status = epics_status;
	}

	/* Compute and set the VAL pvs value from DVAL, DIR, and OFF. */

	dval = *(mr->dval_value);
	off  = *(mr->off_value);

	if ( *(mr->dir_enum_value) != 0 ) {
		dir = -1;		/* Neg */
	} else {
		dir = 1;		/* Pos */
	}

	val = dval * dir + off;

	*(mr->val_value) = val;
	
#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
	MX_DEBUG(-2,("%s: 'VAL = %g, 'DVAL' = %g, DIR = %d, OFF = %g",
		fname, val, dval, dir, off));
#endif

	mr->val_pv->genCopy( mr->val_value );

	mr->val_pv->set_status_severity_and_timestamp( 0, 0 );

	mr->val_pv->post_value_changed();

	return S_casApp_success;
}

caStatus
mxcaMotorDvalPV::write( void )
{
	static const char fname[] = "mxcaMotorDvalPV::write()";

	mxcaMotorRecord *mr;
	MX_RECORD *mx_record;
	MX_MOTOR *mx_motor;
	caStatus epics_status;
	double new_dval, val, raw_set_position;
	int dir;
	short lvio;
	mx_bool_type value_changed;
	mx_status_type mx_status;

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif
	if ( this->get_pv_value_initialized() == FALSE ) {
		this->set_pv_value_initialized( TRUE );
	}

	epics_status = S_casApp_success;

	mr = (mxcaMotorRecord *) this->epicsRecord;

#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
	MX_DEBUG(-2,("%s: PV '%s', *(mr->set_value) = %d",
		fname, this->PVname, *(mr->set_value) ));
#endif

	mx_record = mr->mx_record;

	if ( mx_record == (MX_RECORD *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RECORD pointer for PV '%s' is NULL.",
				this->PVname );

		return S_cas_internal;
	}

	mx_motor = (MX_MOTOR *) mx_record->record_class_struct;

	if ( mx_motor == (MX_MOTOR *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MOTOR pointer for MX record '%s' is NULL.",
			mx_record->name );

		return S_cas_internal;
	}

	new_dval = *(mr->dval_value);

	if ( *(mr->dir_enum_value) != 0 ) {
		dir = -1;		/* Neg */
	} else {
		dir = 1;		/* Pos */
	}

	if ( (new_dval > *(mr->dhlm_value))
	  || (new_dval < *(mr->dllm_value)) )
	{
		lvio = 1;
	} else {
		lvio = 0;
	}

	*(mr->lvio_value) = lvio;

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("%s: PV '%s' LVIO = %hd",
		fname, this->PVname, *(mr->lvio_value) ));
#endif

	mr->lvio_pv->genCopy( mr->lvio_value);

	mr->lvio_pv->post_value_changed();

	/* The DVAL PV does different things depending on the value of
	 * the SET PV found in *(this->set_value).
	 */

	if ( *(mr->set_value) == MXCA_MOTOR_USE_MODE ) {

		/* If the SET PV == "Use" and the SPMG PV == "Go"
		 * and the LVIO PV == 0, then we command a move
		 * through the 'destination' PV.
		 */

		if ( ( *(mr->spmg_value) == MXCA_MOTOR_SPMG_GO )
		  && ( lvio == 0 ) )
		{

#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
			MX_DEBUG(-2,("%s: moving to new destination %g",
				fname, new_dval ));
#endif

			/* The documentation says that DMOV must generate
			 * at least a pulse here, even if the motor has
			 * been commanded to move to its current position.
			 */

			*(mr->dmov_value) = 0;

			mr->dmov_pv->genCopy( mr->dmov_value );
			mr->dmov_pv->post_value_changed();

			/* Update the TDIR PV. */

			if ( new_dval >= mx_motor->position ) {
				*(mr->tdir_value) = 1;
			} else {
				*(mr->tdir_value) = 0;
			}

			mr->tdir_pv->genCopy( mr->tdir_value );

			mr->tdir_pv->post_value_changed();

			/* Setting this flag here makes sure that we see
			 * the end of the pulse in mxcaMotorMstaPV::read().
			 */

			mr->msta_pv->force_value_change = TRUE;

			/* The write() call starts the actual move.
			 * 
			 * It does this by writing to the MX 'destination'
			 * field as set up by mxcaPV::configure_mx_field_pv() 
			 * in our constructor.
			 */

			epics_status = mxcaPV::write();
		}

		*(mr->val_value) = new_dval * dir + *(mr->off_value);

#if MXCA_MOTOR_DEBUG
		MX_DEBUG(-2,("%s(Use): VAL = %g, DVAL = %g, DIR = %d, OFF = %g",
			this->PVname, *(mr->val_value),
			new_dval, dir, *(mr->off_value) ));
#endif

		/* Send out the new value of VAL. */

		mr->val_pv->genCopy( mr->val_value );
		mr->val_pv->post_value_changed();

	} else {
		/* Otherwise, we interpret the 'write' request as a 
		 * request to redefine the current position.
		 */


#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
		MX_DEBUG(-2,("%s: PV '%s' invoked with SET != 0",
			fname, this->PVname));
#endif

#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
		MX_DEBUG(-2,("%s: PV '%s', redefining dial position to %g",
			fname, this->PVname, new_dval ));
#endif

#if MXCA_MOTOR_DEBUG
		MX_DEBUG(-2,("%s: FOFF = %d", fname, *(mr->foff_value) ));
#endif

		/*-----------------------------------------------------*/

		/* If FOFF is set to 0 (Variable), then all we do is to
		 * change the value of OFF.
		 */

		if ( *(mr->foff_value) == MXCA_MOTOR_FOFF_VARIABLE ) {
			val = *(mr->val_value);

#if MXCA_MOTOR_DEBUG
			MX_DEBUG(-2,("%s: VAL = %g, DVAL = %g, DIR = %d",
			fname, val, new_dval, dir));
#endif

			*(mr->off_value) = val - new_dval * dir;

#if MXCA_MOTOR_DEBUG
			MX_DEBUG(-2,("%s: OFF = %g", fname, *(mr->off_value) ));
#endif

			mr->off_pv->genCopy( mr->off_value );

			mr->off_pv->post_value_changed();

			this->post_value_changed();

			return S_casApp_success;
		}

		/*-----------------------------------------------------*/

		/* FOFF is not 0, so we will redefine the MX position. */

		/* Copy the new value to the MX 'set_position' record field. */

		mx_motor->set_position = new_dval;

		raw_set_position = mx_divide_safely(
				mx_motor->set_position - mx_motor->offset,
					mx_motor->scale );

		if ( mx_motor->subclass == MXC_MTR_STEPPER ) {
			mx_motor->raw_set_position.stepper =
				mx_round( raw_set_position );
		} else {
			mx_motor->raw_set_position.analog = raw_set_position;
		}

		/* Process the MX 'set_position' record field, which 
		 * redefines the position.
		 */

		if (this->set_position_rf == (MX_RECORD_FIELD *) NULL) {
			(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
				"The 'set_position' record field pointer for "
				"PV '%s' is NULL.", this->PVname );

			return S_cas_internal;
		}

		mx_status = mx_process_record_field( mx_record,
						this->set_position_rf,
						MX_PROCESS_PUT,
						&value_changed );

		if ( mx_status.code != MXE_SUCCESS ) {
			return S_cas_internal;
		}

		if ( value_changed ) {
			this->post_value_changed();
		}
	}

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return epics_status;
}

mx_bool_type
mxcaMotorDvalPV::post_value_changed( void )
{
#if MXCA_MOTOR_DEBUG_DIAL_POST_VALUE_CHANGED
	static const char fname[] = "mxcaMotorDvalPV::post_value_changed()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	mxcaMotorRecord *mr;

	mxcaPV::post_value_changed();

	mr = (mxcaMotorRecord *) this->epicsRecord;

#if MXCA_MOTOR_DEBUG_DIAL_POST_VALUE_CHANGED
	MX_DEBUG(-2,("%s: mr->val_pv = %p", fname, mr->val_pv));
#endif

	if ( mr->val_pv != NULL ) {
		mr->val_pv->post_value_changed();
	}

	return TRUE;
}

/*-----------------------------------------------------------------------*/

mxcaMotorValPV::mxcaMotorValPV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, ".VAL" )
{
#if MXCA_MOTOR_DEBUG
	static const char fname[] = "mxcaMotorValPV::mxcaMotorValPV()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	mxcaMotorRecord *mr;

	/* Setup the this->mx_value data structure. */

	double double_value = 0.0;

	this->configure_internal_pv( aitEnumFloat64, 0, NULL, &double_value );

	/* NOTE: The pointers for val_pv and val_value in the mxcaMotorDvalPV
	 * object are set up by the mxcaMotorValPV constructor, since this
	 * object will be created after the mxcaMotorValPV class's constructor
	 * runs.
	 */

	mr = (mxcaMotorRecord *) this->epicsRecord;

	mr->val_value = &(this->val_value);

	this->mx_value = mr->val_value;

	/*---*/

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return;
}

caStatus
mxcaMotorValPV::read( void )
{
	mxcaMotorRecord *mr;

	mr = (mxcaMotorRecord *) this->epicsRecord;

	mr->dval_pv->read();

	return S_casApp_success;
}

caStatus
mxcaMotorValPV::write( void )
{
#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
	static const char fname[] = "mxcaMotorValPV::write( void )";
#endif

	mxcaMotorRecord *mr;
	double dval, off, val;
	int dir;

#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	mr = (mxcaMotorRecord *) this->epicsRecord;

	/* If we are in Set mode and the offset is variable, then
	 * pass on the request to the DVAL PV object, without doing
	 * anything else beforehand.
	 */

	if ( ( *(mr->set_value) == MXCA_MOTOR_SET_MODE )
	  && ( *(mr->foff_value) == MXCA_MOTOR_FOFF_VARIABLE) )
	{
		mr->dval_pv->write();

		return S_casApp_success;
	}

	/* Otherwise, we compute a new value for DVAL before calling
	 * the DVAL write() method.
	 */

	val = *(mr->val_value);
	off = *(mr->off_value);

	if ( *(mr->dir_enum_value) != 0 ) {
		dir = -1;		/* Neg */
	} else {
		dir = 1;		/* Pos */
	}

	dval = dir * ( val - off );

	*(mr->dval_value) = dval;

#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
	MX_DEBUG(-2,("%s: 'VAL = %g, 'DVAL' = %g, DIR = %d, OFF = %g",
		fname, val, dval, dir, off));
#endif
	mr->dval_pv->genCopy( mr->dval_value );

	mr->dval_pv->write();

#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return S_casApp_success;
}

/*-----------------------------------------------------------------------*/

mxcaMotorRlvPV::mxcaMotorRlvPV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, ".RLV" )
{
	this->rlv_value = 0.0;

	this->mx_value = &(this->rlv_value);

	this->configure_internal_pv( aitEnumFloat64,
					0, NULL, &(this->rlv_value) );
}

caStatus
mxcaMotorRlvPV::read( void )
{
	return S_casApp_success;
}

caStatus
mxcaMotorRlvPV::write( const gdd &gdd_object )
{
	mxcaMotorRecord *mr;
	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();
	double old_val, new_val;

	/* Get the value from EPICS. */
	
	table.smartCopy( this->epics_value, &gdd_object );

	gdd_object.getConvert( this->rlv_value );

	/* Add the value of RLV to the existing value of VAL. */

	mr = (mxcaMotorRecord *) this->epicsRecord;

	old_val = *(mr->val_value);

	new_val = old_val + this->rlv_value;

	*(mr->val_value) = new_val;

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("'%s' RLV = %g, old VAL = %g, new VAL = %g",
		this->PVname, this->rlv_value, old_val, new_val ));
#endif

	/* Start the move. */

	mr->val_pv->write();

	/* Reset RLV to 0. */

	this->rlv_value = 0.0;

	this->genCopy( &(this->rlv_value) );

	this->post_value_changed();

	return S_casApp_success;
}

/*-----------------------------------------------------------------------*/

mxcaMotorDialPV::mxcaMotorDialPV( mxcaRecord *epics_record,
				const char *epics_field_name,
				const char *mx_field_name,
				mx_bool_type use_monitor_deadband ) :
			mxcaPV( epics_record, epics_field_name )
{
#if MXCA_MOTOR_DEBUG
	static const char fname[] = "mxcaMotorDialPV::mxcaMotorDialPV()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
	MX_DEBUG(-2,("%s: this = %p", fname, this));
#endif

	mxcaMotorRecord *mr;

	mr = (mxcaMotorRecord *) epics_record;

	this->configure_mx_field_pv( mx_field_name );

	this->dial_value = (double *) this->mx_value;

	this->user_value = NULL;
	this->user_pv    = NULL;

	this->use_monitor_deadband = use_monitor_deadband;
	this->old_position = DBL_MAX;

	if ( strcmp( epics_field_name, ".DHLM" ) == 0 ) {
		mr->dhlm_value = this->dial_value;

#if MXCA_MOTOR_DEBUG
		MX_DEBUG(-2,("%s: PV '%s', mr->dhlm_value = %p",
			fname, this->PVname, mr->dhlm_value));
#endif
	} else
	if ( strcmp( epics_field_name, ".DLLM" ) == 0 ) {
		mr->dllm_value = this->dial_value;

#if MXCA_MOTOR_DEBUG
		MX_DEBUG(-2,("%s: PV '%s', mr->dllm_value = %p",
			fname, this->PVname, mr->dllm_value));
#endif
	}
}

caStatus
mxcaMotorDialPV::read( void )
{
#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
	static const char fname[] = "mxcaMotorDialPV::read( void )";
#endif

	mxcaMotorRecord *mr;
	double dial, off, user;
	int dir;
	caStatus epics_status;

	epics_status = mxcaPV::read();

#if 1
	/* Shut up GCC 4.6 unused but set warning. */

	epics_status = epics_status;
#endif

	dial = *(this->dial_value);

	mr = (mxcaMotorRecord *) this->epicsRecord;

	off  = *(mr->off_value);

	if ( *(mr->dir_enum_value) != 0 ) {
		dir = -1;		/* Neg */
	} else {
		dir = 1;		/* Pos */
	}
	
	user = dial * dir + off;

	*(this->user_value) = user;

#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
	MX_DEBUG(-2,("%s: PV '%s', USER = %g, DIAL = %g, DIR = %d, OFF = %g",
		fname, this->PVname, user, dial, dir, off));
#endif

	this->user_pv->genCopy( this->user_value );

	this->user_pv->set_status_severity_and_timestamp( 0, 0 );

	if ( this->use_monitor_deadband )
	{
		double difference = fabs( user - this->old_position );

		MX_MOTOR *motor = (MX_MOTOR *)
				this->mx_record->record_class_struct;

		if ( difference > (motor->value_change_threshold) ) {
			this->old_position = user;

			this->user_pv->post_value_changed();
		}
	}

	return S_casApp_success;
}

caStatus
mxcaMotorDialPV::write( void )
{
	caStatus epics_status;

	epics_status = mxcaPV::write();

	return epics_status;
}

mx_bool_type
mxcaMotorDialPV::post_value_changed( void )
{
#if MXCA_MOTOR_DEBUG_DIAL_POST_VALUE_CHANGED
	static const char fname[] = "mxcaMotorDialPV::post_value_changed()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	mxcaPV::post_value_changed();

#if MXCA_MOTOR_DEBUG_DIAL_POST_VALUE_CHANGED
	MX_DEBUG(-2,("%s: this->user_pv = %p", fname, this->user_pv));
#endif

	if ( this->user_pv != NULL ) {
		this->user_pv->post_value_changed();
	}

	return TRUE;
}

/*-----------------------------------------------------------------------*/

mxcaMotorUserPV::mxcaMotorUserPV( mxcaRecord *epics_record,
				const char *epics_field_name,
				mxcaMotorDialPV *dial_pv ) :
			mxcaPV( epics_record, epics_field_name )
{
#if MXCA_MOTOR_DEBUG
	static const char fname[] = "mxcaMotorUserPV::mxcaMotorUserPV()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
	MX_DEBUG(-2,("%s: this = %p", fname, this));
	MX_DEBUG(-2,("%s: dial_pv = %p", fname, dial_pv));
#endif

	double double_value;

	double_value = 0.0;
	this->configure_internal_pv( aitEnumFloat64, 0, NULL, &double_value );

	this->user_value = new double;

	this->mx_value = this->user_value;

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("%s: PV '%s' epics_value = %p, mx_value = %p",
		fname, this->PVname, this->epics_value, this->mx_value));
#endif

	this->dial_pv = dial_pv;
	this->dial_value = dial_pv->dial_value;

	dial_pv->user_pv = this;
	dial_pv->user_value = this->user_value;

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("%s: dial_pv = %p, this = %p, dial_pv->user_pv = %p",
		fname, dial_pv, this, dial_pv->user_pv));
#endif
}

caStatus
mxcaMotorUserPV::read( void )
{
	this->dial_pv->read();

	return S_casApp_success;
}

caStatus
mxcaMotorUserPV::write( void )
{
#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
	static const char fname[] = "mxcaMotorUserPV::write( void )";
#endif

	mxcaMotorRecord *mr;
	double dial, off, user;
	int dir;

	user = *(this->user_value);

	mr = (mxcaMotorRecord *) this->epicsRecord;

	off  = *(mr->off_value);

	if ( *(mr->dir_enum_value) != 0 ) {
		dir = -1;		/* Neg */
	} else {
		dir = 1;		/* Pos */
	}

	dial = dir * ( user - off );

	*(this->dial_value) = dial;

#if MXCA_MOTOR_DEBUG_USER_AND_DIAL
	MX_DEBUG(-2,("%s: PV '%s', USER = %g, DIAL = %g, DIR = %d, OFF = %g",
		fname, this->PVname, user, dial, dir, off));
#endif

	this->genCopy( this->dial_value );

	this->dial_pv->write();

	this->dial_pv->post_value_changed();

	return S_casApp_success;
}

/*-----------------------------------------------------------------------*/

mxcaMotorRrbvPV::mxcaMotorRrbvPV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, ".RRBV" )
{
	this->configure_mx_field_pv( "raw_position" );

	return;
}

mx_bool_type
mxcaMotorRrbvPV::post_value_changed( void )
{
#if MXCA_MOTOR_DEBUG
	static const char fname[] = "mxcaMotorRrbvPV::post_value_changed()";
#endif

	mxcaMotorRecord *mr;
	MX_MOTOR *motor;
	const char *epics_record_name;
	double raw_position;

	mr = (mxcaMotorRecord *) this->epicsRecord;

	motor = (MX_MOTOR *) mr->mx_record->record_class_struct;

	epics_record_name = this->epicsRecord->get_epics_record_name();

#if 1
	/* Shut up GCC 4.6 unused but set warning. */

	epics_record_name = epics_record_name;
#endif

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("%s invoked for EPICS record '%s'.",
		fname, epics_record_name ));
#endif

	/* Post value changed for RRBV. */

	mxcaPV::post_value_changed();

	/* Compute new values for DVAL and VAL. */

	switch( motor->subclass ) {
	case MXC_MTR_STEPPER:
		raw_position = motor->raw_position.stepper;
		break;
	case MXC_MTR_ANALOG:
		raw_position = motor->raw_position.analog;
		break;
	}

#if 1
	/* Shut up GCC 4.6 unused but set warning. */

	raw_position = raw_position;
#endif

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("%s: '%s' raw_position = %f",
		fname, epics_record_name, raw_position));
#endif

	/* Post value changed for DRBV.
	 *
	 * The post_value_changed() handler for DRBV will, in turn, post
	 * value changed for RBV, so we should not do it ourself here.
	 */

	*(mr->drbv_pv->dial_value) = motor->position;

	mr->drbv_pv->post_value_changed();

	return TRUE;
}

/*-----------------------------------------------------------------------*/

mxcaMotorDirPV::mxcaMotorDirPV( mxcaRecord *epics_record ) :
			mxcaEnumPV( epics_record, ".DIR" )
{
	aitEnum16 enum_value;

	enum_value = 0;
	this->configure_enum_pv( enum_value, 2, "Pos", "Neg" );

	return;
}

caStatus
mxcaMotorDirPV::write( const gdd &gdd_object )
{
	caStatus epics_status;

	mxcaMotorRecord *mr;
	int dir;
	double off;

	epics_status = mxcaEnumPV::write( gdd_object );

	if ( epics_status != S_casApp_success )
		return epics_status;

	/*---*/

	if ( this->enum_value != 0 ) {
		dir = -1;
	} else {
		dir = 1;
	}

	mr = (mxcaMotorRecord *) this->epicsRecord;

	off = *(mr->off_value);

	/*---*/

	*(mr->val_value) = off + dir * *(mr->dval_value);

	mr->val_pv->genCopy( mr->val_value );

	/*---*/

	*(mr->hlm_pv->user_value) = off + dir * *(mr->dhlm_pv->dial_value);

	mr->hlm_pv->genCopy( mr->hlm_pv->user_value );

	/*---*/

	*(mr->llm_pv->user_value) = off + dir * *(mr->dllm_pv->dial_value);

	mr->llm_pv->genCopy( mr->llm_pv->user_value );

	/*---*/

	*(mr->rbv_pv->user_value) = off + dir * *(mr->drbv_pv->dial_value);

	mr->rbv_pv->genCopy( mr->rbv_pv->user_value );

	/*---*/

	this->post_value_changed();

	return S_casApp_success;
}

mx_bool_type
mxcaMotorDirPV::post_value_changed( void )
{
	mxcaMotorRecord *mr;

	mxcaPV::post_value_changed();

	mr = (mxcaMotorRecord *) this->epicsRecord;

	if ( mr->val_pv != NULL ) {
		mr->val_pv->post_value_changed();
	}
	if ( mr->hlm_pv != NULL ) {
		mr->hlm_pv->post_value_changed();
	}
	if ( mr->llm_pv != NULL ) {
		mr->llm_pv->post_value_changed();
	}
	if ( mr->rbv_pv != NULL ) {
		mr->rbv_pv->post_value_changed();
	}

	return TRUE;
}

/*-----------------------------------------------------------------------*/

mxcaMotorOffPV::mxcaMotorOffPV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, ".OFF" )
{
	double double_value;
	double *double_ptr;

	double_value = 0.0;
	this->configure_internal_pv( aitEnumFloat64, 0, NULL, &double_value );

	double_ptr = new double;
	*double_ptr = 0.0;
	this->set_mx_value_pointer( double_ptr );

	return;
}

caStatus
mxcaMotorOffPV::write( void )
{
	caStatus epics_status;

	mxcaMotorRecord *mr;
	int dir;
	double off;

	epics_status = mxcaPV::write();

	if ( epics_status != S_casApp_success )
		return epics_status;

	/*---*/

	mr = (mxcaMotorRecord *) this->epicsRecord;

	if ( mr->dir_pv->enum_value != 0 ) {
		dir = -1;
	} else {
		dir = 1;
	}

	off = *( mr->off_value );

	/*---*/

	*(mr->val_value) = off + dir * *(mr->dval_value);

	mr->val_pv->genCopy( mr->val_value );

	/*---*/

	*(mr->hlm_pv->user_value) = off + dir * *(mr->dhlm_pv->dial_value);

	mr->hlm_pv->genCopy( mr->hlm_pv->user_value );

	/*---*/

	*(mr->llm_pv->user_value) = off + dir * *(mr->dllm_pv->dial_value);

	mr->llm_pv->genCopy( mr->llm_pv->user_value );

	/*---*/

	*(mr->rbv_pv->user_value) = off + dir * *(mr->drbv_pv->dial_value);

	mr->rbv_pv->genCopy( mr->rbv_pv->user_value );

	/*---*/

	this->post_value_changed();

	return S_casApp_success;
}

mx_bool_type
mxcaMotorOffPV::post_value_changed( void )
{
	mxcaMotorRecord *mr;

	mxcaPV::post_value_changed();

	mr = (mxcaMotorRecord *) this->epicsRecord;

	if ( mr->val_pv != NULL ) {
		mr->val_pv->post_value_changed();
	}
	if ( mr->hlm_pv != NULL ) {
		mr->hlm_pv->post_value_changed();
	}
	if ( mr->llm_pv != NULL ) {
		mr->llm_pv->post_value_changed();
	}
	if ( mr->rbv_pv != NULL ) {
		mr->rbv_pv->post_value_changed();
	}

	return TRUE;
}

/*-----------------------------------------------------------------------*/

mxcaMotorDiffPV::mxcaMotorDiffPV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, ".DIFF" )
{
	mxcaMotorRecord *mr;
	MX_MOTOR *motor;
	double difference;

	mr = (mxcaMotorRecord *) this->epicsRecord;

	motor = (MX_MOTOR *) mr->mx_record->record_class_struct;

	difference = 0.0;
	this->configure_internal_pv( aitEnumFloat64, 0, NULL, &difference );

	this->mx_value = NULL;

	this->old_difference = DBL_MAX;

	switch( motor->subclass ) {
	case MXC_MTR_STEPPER:
		this->deadband = motor->raw_move_deadband.stepper;
		break;
	case MXC_MTR_ANALOG:
		this->deadband = motor->raw_move_deadband.analog;
		break;
	}

	return;
}

caStatus
mxcaMotorDiffPV::read( void )
{
	mxcaMotorRecord *mr;
	MX_MOTOR *motor;
	double difference, delta_difference;

	mr = (mxcaMotorRecord *) this->epicsRecord;

	motor = (MX_MOTOR *) mr->mx_record->record_class_struct;

	/* Get the difference between the current position
	 * and the destination.
	 */

	difference = motor->destination - motor->position;

	this->genCopy( &difference );

	delta_difference = fabs( difference - this->old_difference );

#if 0
	MX_DEBUG(-2,("'%s' difference = %g, delta_difference = %g",
		this->PVname, difference, delta_difference));
#endif

	if ( delta_difference > this->deadband ) {
		this->old_difference = difference;

		this->post_value_changed();
	}
	
	return S_casApp_success;
}

/*-----------------------------------------------------------------------*/

mxcaMotorSpmgPV::mxcaMotorSpmgPV( mxcaRecord *epics_record ) :
			mxcaEnumPV( epics_record, ".SPMG" )
{
	mxcaMotorRecord *mr;
	aitEnum16 enum_value;

	enum_value = MXCA_MOTOR_SPMG_GO;

	this->configure_enum_pv( enum_value, 4, "Stop", "Pause", "Move", "Go" );

	mr = (mxcaMotorRecord *) this->epicsRecord;

	mr->spmg_value = &(this->enum_value);

	this->mx_value = mr->spmg_value;

	return;
}

caStatus
mxcaMotorSpmgPV::write( const gdd &gdd_object )
{
	mxcaMotorRecord *mr;
	MX_MOTOR *motor;
	mx_status_type mx_status;

	mxcaEnumPV::write( gdd_object );

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("SPMG: enum_value = %d", this->enum_value));
#endif

	mr = (mxcaMotorRecord *) this->epicsRecord;

	motor = (MX_MOTOR *) this->mx_record->record_class_struct;

	if ( this->enum_value == MXCA_MOTOR_SPMG_STOP ) {

		mx_status = mx_motor_soft_abort( mr->mx_record );

		if ( mx_status.code != MXE_SUCCESS )
			return S_casApp_undefined;	/* FIXME - Exception */
		
		mx_status = mx_wait_for_motor_stop( mr->mx_record, 0 );

		if ( mx_status.code != MXE_SUCCESS )
			return S_casApp_undefined;	/* FIXME - Exception */

		mr->set_dval_to_drbv();
	} else
	if ( this->enum_value == MXCA_MOTOR_SPMG_MOVE ) {
		/* Command a move to the currently set destination. */

		mx_status = mx_motor_move_absolute( mr->mx_record,
						motor->destination, 0 );
	}

	return S_casApp_success;
}

/*-----------------------------------------------------------------------*/

mxcaMotorJogPV::mxcaMotorJogPV( mxcaRecord *epics_record,
				const char *epics_field_name,
				int direction ) :
			mxcaPV( epics_record, epics_field_name )
{
	short short_value;

	short_value = 0;
	this->configure_internal_pv( aitEnumInt16, 0, NULL, &short_value );

	this->direction = direction;
	this->mx_value = &(this->jog_value);

	return;
}

caStatus
mxcaMotorJogPV::write( const gdd &gdd_object )
{
	mxcaMotorRecord *mr;
	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();
	double position;
	mx_status_type mx_status;

	/* Copy from the object sent by the client. */

	table.smartCopy( this->epics_value, &gdd_object );

	gdd_object.getConvert( this->jog_value );

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("'%s' jog_value = %d", this->PVname, this->jog_value));
#endif

	mr = (mxcaMotorRecord *) this->epicsRecord;

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("JOG: DMOV = %hd", *(mr->dmov_value) ));
#endif

	mr->msta_pv->force_value_change = TRUE;

	if ( this->jog_value ) {
		/* Starting a jog. (JOGx == 1) */

		if ( *(mr->dmov_value) != 1 ) {
			mx_status = mx_motor_soft_abort( mr->mx_record );

			if ( mx_status.code != MXE_SUCCESS )
				return S_casApp_undefined;  /* FIXME */

			mx_status = mx_wait_for_motor_stop( mr->mx_record, 0 );

			if ( mx_status.code != MXE_SUCCESS )
				return S_casApp_undefined;  /* FIXME */

			mr->set_dval_to_drbv();
		}

		mx_status = mx_motor_constant_velocity_move( mr->mx_record,
							this->direction );

		if ( mx_status.code != MXE_SUCCESS )
			return S_casApp_undefined;  /* FIXME */
	} else {
		/* Ending a jog. (JOGx == 0) */

		/* Get the current motor position from MX. */

		mx_status = mx_motor_get_position( mr->mx_record, &position );

		if ( mx_status.code != MXE_SUCCESS )
			return S_casApp_undefined;  /* FIXME */

		/* Stop the motor. */

		mx_status = mx_motor_soft_abort( mr->mx_record );

		if ( mx_status.code != MXE_SUCCESS )
			return S_casApp_undefined;  /* FIXME */

		mx_status = mx_wait_for_motor_stop( mr->mx_record, 0 );

		if ( mx_status.code != MXE_SUCCESS )
			return S_casApp_undefined;  /* FIXME */

		/* Finish by commanding a move to the position captured above.*/

		mx_status = mx_motor_move_absolute( mr->mx_record,
						position, 0 );

		if ( mx_status.code != MXE_SUCCESS )
			return S_casApp_undefined;  /* FIXME */
	}

	return S_casApp_success;
}

/*-----------------------------------------------------------------------*/

mxcaMotorHomePV::mxcaMotorHomePV( mxcaRecord *epics_record,
				const char *epics_field_name,
				int direction ) :
			mxcaPV( epics_record, epics_field_name )
{
	short short_value;

	short_value = 0;
	this->configure_internal_pv( aitEnumInt16, 0, NULL, &short_value );

	this->direction = direction;
	this->mx_value = &(this->home_value);

	return;
}

caStatus
mxcaMotorHomePV::read( void )
{
	mxcaMotorRecord *mr;
	MX_MOTOR *motor;
	unsigned long motor_status;
	mx_status_type mx_status;

	mr = (mxcaMotorRecord *) this->epicsRecord;

	if ( this->get_pv_value_initialized() == FALSE ) {
		this->set_pv_value_initialized( TRUE );
	}

	motor = (MX_MOTOR *) mr->mx_record->record_class_struct;

	mx_status = mx_motor_get_status( mr->mx_record, &motor_status );

	if ( mx_status.code != MXE_SUCCESS ) {
		return S_casApp_undefined;	/* FIXME */
	}

	if ( motor->home_search_in_progress ) {
		this->home_value = 1;
	} else {
		this->home_value = 0;
	}

#if 0
	MX_DEBUG(-2,("read: status = %#lx, ip = %d, home_value = %d",
	    motor_status, motor->home_search_in_progress, this->home_value ));
#endif

	this->genCopy( &(this->home_value) );

	return S_casApp_success;
}

caStatus
mxcaMotorHomePV::write( const gdd &gdd_object )
{
	mxcaMotorRecord *mr;
	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();
	mx_status_type mx_status;

	/* Copy from the object sent by the client. */

	table.smartCopy( this->epics_value, &gdd_object );

	gdd_object.getConvert( this->home_value );

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("'%s' home_value = %d", this->PVname, this->home_value));
#endif

	mr = (mxcaMotorRecord *) this->epicsRecord;

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("HOME: DMOV = %hd", *(mr->dmov_value) ));
#endif

	if ( this->get_pv_value_initialized() == FALSE ) {
		this->set_pv_value_initialized( TRUE );
	}

	if ( this->home_value == 1 ) {
		/* Starting a home search. (HOMEx == 1) */

		if ( *(mr->dmov_value) != 1 ) {

			mx_status = mx_motor_soft_abort( mr->mx_record );

			if ( mx_status.code != MXE_SUCCESS )
				return S_casApp_undefined;  /* FIXME */

			mx_status = mx_wait_for_motor_stop( mr->mx_record, 0 );

			if ( mx_status.code != MXE_SUCCESS )
				return S_casApp_undefined;  /* FIXME */

			mr->set_dval_to_drbv();
		}

		mr->msta_pv->force_value_change = TRUE;

#if 0
		MX_DEBUG(-2,("write: BEFORE HOME"))
#endif

		mx_status = mx_motor_home_search( mr->mx_record,
						this->direction, 0 );

#if 0
		MX_DEBUG(-2,("write: AFTER HOME"))
#endif

		if ( mx_status.code != MXE_SUCCESS )
			return S_casApp_undefined;  /* FIXME */
	} else {
		/* All other values for HOMEx are ignored. */
	}

	return S_casApp_success;
}

/*-----------------------------------------------------------------------*/

mxcaMotorTweakPV::mxcaMotorTweakPV( mxcaRecord *epics_record,
				const char *epics_field_name,
				int direction ) :
			mxcaPV( epics_record, epics_field_name )
{
	this->direction = direction;

	this->tweak_value = 0;
	this->configure_internal_pv( aitEnumInt16,
					0, NULL, &(this->tweak_value) );
	this->mx_value = &(this->tweak_value);

	return;
}

caStatus
mxcaMotorTweakPV::write( const gdd &gdd_object )
{
	mxcaMotorRecord *mr;
	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();
	double tweak, old_val, new_val;
	mx_status_type mx_status;

	/* Copy from the object sent by the client. */

	table.smartCopy( this->epics_value, &gdd_object );

	gdd_object.getConvert( this->tweak_value );

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("'%s' tweak_value = %d", this->PVname, this->tweak_value));
#endif

	mr = (mxcaMotorRecord *) this->epicsRecord;

#if MXCA_MOTOR_DEBUG
	MX_DEBUG(-2,("Tweak: DMOV = %hd", *(mr->dmov_value) ));
#endif

	if ( this->tweak_value == 1 ) {
		/* Begin a tweak. */

		if ( *(mr->dmov_value) != 1 ) {
			mx_status = mx_motor_soft_abort( mr->mx_record );

			if ( mx_status.code != MXE_SUCCESS )
				return S_casApp_undefined;  /* FIXME */

			mx_status = mx_wait_for_motor_stop( mr->mx_record, 0 );

			if ( mx_status.code != MXE_SUCCESS )
				return S_casApp_undefined;  /* FIXME */

			mr->set_dval_to_drbv();
		}

		/* Add or subtract the tweak from the existing value of VAL. */

		tweak = *(mr->twv_value) * (this->direction);

		old_val = *(mr->val_value);

		new_val = old_val + tweak;

		*(mr->val_value) = new_val;

#if MXCA_MOTOR_DEBUG
		MX_DEBUG(-2,("'%s' Tweak = %g, old VAL = %g, new VAL = %g",
			this->PVname, tweak, old_val, new_val));
#endif

		/* Start the move. */

		mr->val_pv->write();

	} else {
		/* All other values for Tweak PVs are ignored. */
	}

	/* Reset the Tweak PV to 0. */

	this->tweak_value = 0;

	this->genCopy( &(this->tweak_value) );

	this->post_value_changed();

	return S_casApp_success;
}

/*-----------------------------------------------------------------------*/

mxcaMotorRxpPV::mxcaMotorRxpPV( mxcaRecord *epics_record,
				const char *epics_field_name ) :
			mxcaPV( epics_record, epics_field_name )
{
	mxcaMotorRecord *mr;
	MX_MOTOR *motor;

	mr = (mxcaMotorRecord *) this->epicsRecord;

	motor = (MX_MOTOR *) mr->mx_record->record_class_struct;

	/* Set the old position to an extremely large value. */

	this->position = DBL_MAX;
	this->old_position = DBL_MAX;

	this->configure_internal_pv( aitEnumFloat64,
					0, NULL, &(this->position) );

	this->mx_value = &(this->position);

	switch( motor->subclass ) {
	case MXC_MTR_STEPPER:
		this->deadband = motor->raw_move_deadband.stepper;
		break;
	case MXC_MTR_ANALOG:
		this->deadband = motor->raw_move_deadband.analog;
		break;
	}

	return;
}

caStatus
mxcaMotorRxpPV::read( void )
{
	mxcaMotorRecord *mr;
	MX_MOTOR *motor;
	double new_position, old_position, difference;

	mr = (mxcaMotorRecord *) this->epicsRecord;

	motor = (MX_MOTOR *) mr->mx_record->record_class_struct;

	/* Get the new raw position of the motor. */

	old_position = this->old_position;

	switch( motor->subclass ) {
	case MXC_MTR_STEPPER:
		new_position = motor->raw_position.stepper;
		break;
	case MXC_MTR_ANALOG:
		new_position = motor->raw_position.analog;
		break;
	}

	difference = fabs( new_position - old_position );

	this->position = new_position;

	this->genCopy( &(this->position) );

	if ( difference > this->deadband ) {
		this->old_position = new_position;

		this->post_value_changed();
	}
	
	return S_casApp_success;
}

/*-----------------------------------------------------------------------*/

/* Note: In MX, (motor->status & MXSF_MTR_IS_BUSY) should not go to zero
 * during a continuous sequence of motions even if the motor speed goes 
 * to zero momentarily during the sequence.  Because of this, in our
 * implementation DMOV and MOVN _are_ inverses of each other, except for
 * the momentary DMOV pulse at the start of a commanded move.
 */

mxcaMotorMstaPV::mxcaMotorMstaPV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, ".MSTA" )
{
	mxcaMotorRecord *mr;

	this->msta_value = 0x2;
	this->athm_value = 0;
	this->dmov_value = 1;
	this->movn_value = 0;

	this->force_value_change = FALSE;

	mr = (mxcaMotorRecord *) this->epicsRecord;

	mr->msta_value = &(this->msta_value);
	mr->athm_value = &(this->athm_value);
	mr->dmov_value = &(this->dmov_value);
	mr->movn_value = &(this->movn_value);

	this->mx_value = &(this->msta_value);

	this->configure_internal_pv( aitEnumInt32,
					0, NULL, &(this->msta_value) );

	/*---*/

	mr->athm_pv = new mxcaPV( mr, ".ATHM" );
	mr->athm_pv->configure_internal_pv( aitEnumInt16,
					0, NULL, &(this->athm_value) );
	mr->athm_pv->set_mx_value_pointer( mr->athm_value );
	mr->athm_pv->set_pv_value_initialized( TRUE );

	/*---*/

	mr->dmov_pv = new mxcaPV( mr, ".DMOV" );
	mr->dmov_pv->configure_internal_pv( aitEnumInt16,
					0, NULL, &(this->dmov_value) );
	mr->dmov_pv->set_mx_value_pointer( mr->dmov_value );
	mr->dmov_pv->set_pv_value_initialized( TRUE );

	/*---*/

	mr->movn_pv = new mxcaPV( mr, ".MOVN" );
	mr->movn_pv->configure_internal_pv( aitEnumInt16,
					0, NULL, &(this->movn_value) );
	mr->movn_pv->set_mx_value_pointer( mr->movn_value );
	mr->movn_pv->set_pv_value_initialized( TRUE );

	/*---*/

	this->motor = (MX_MOTOR *) mr->mx_record->record_class_struct;

	this->motor->status = 0xffffffff;

	this->status_rf = mx_get_record_field( mr->mx_record, "status" );
}

caStatus
mxcaMotorMstaPV::read( void )
{
	mxcaMotorRecord *mr;
	unsigned long motor_status;
	mx_bool_type value_changed;
	mx_status_type mx_status;

	mr = (mxcaMotorRecord *) this->epicsRecord;

	mx_status = mx_process_record_field( mx_record,
					this->status_rf,
					MX_PROCESS_GET,
					&value_changed );

	if ( mx_status.code != MXE_SUCCESS ) 
		return S_casApp_undefined;	/* FIXME */

	/* All of the MX status information is found in motor->status. */

	motor_status = motor->status;

	/* Make sure that we see the end of any pulse started
	 * by mxcaMotorDvalPV::write().
	 */

	if ( this->force_value_change ) {
		this->force_value_change = FALSE;

		if ( ( motor_status & MXSF_MTR_IS_BUSY ) == FALSE ) {
			value_changed = TRUE;
		}
	}

	/* Compute all the status variables. */

	this->msta_value = 0;

	if ( motor_status & MXSF_MTR_IS_BUSY ) {
		this->dmov_value = 0;
		this->movn_value = 1;
		this->msta_value &= ~(0x2);
		this->msta_value |= 0x400;
	} else {
		this->dmov_value = 1;
		this->movn_value = 0;
		this->msta_value |= 0x2;
		this->msta_value &= ~(0x400);
	}

	if ( motor_status & MXSF_MTR_POSITIVE_LIMIT_HIT ) {
		this->msta_value |= 0x4;
	} else {
		this->msta_value &= ~(0x4);
	}

	if ( motor_status & MXSF_MTR_NEGATIVE_LIMIT_HIT ) {
		this->msta_value |= 0x2000;
	} else {
		this->msta_value &= ~(0x2000);
	}

	if ( motor_status & MXSF_MTR_HOME_SEARCH_SUCCEEDED ) {
		this->athm_value = 1;
		this->msta_value |= 0x4088;
	} else {
		this->athm_value = 0;
		this->msta_value &= ~(0x4088);
	}

	if ( motor_status & MXSF_MTR_FOLLOWING_ERROR ) {
		this->msta_value |= 0x40;
	} else {
		this->msta_value &= ~(0x40);
	}

	if ( ( motor_status & MXSF_MTR_DRIVE_FAULT )
	  || ( motor_status & MXSF_MTR_AXIS_DISABLED ) )
	{
		this->msta_value |= 0x200;
	} else {
		this->msta_value &= ~(0x200);
	}

	if ( motor_status & MXSF_MTR_OPEN_LOOP ) {
		this->msta_value &= ~(0x20);
	} else {
		this->msta_value |= 0x20;
	}

#if MXCA_MOTOR_DEBUG_MSTA
	MX_DEBUG(-2,
	("PV '%s', status = %#lx, MSTA = %#lx, DMOV = %hd, ATHM = %hd",
		this->PVname, motor_status,
		this->msta_value, this->dmov_value, this->athm_value ));
#endif

	this->genCopy( &(this->msta_value) );

	mr->athm_pv->genCopy( &(this->athm_value) );
	mr->dmov_pv->genCopy( &(this->dmov_value) );
	mr->movn_pv->genCopy( &(this->movn_value) );

	if ( value_changed ) {
		this->post_value_changed();
		mr->athm_pv->post_value_changed();
		mr->dmov_pv->post_value_changed();
		mr->movn_pv->post_value_changed();
	}

	return S_casApp_success;
}

