/*
 * Name:    mxcaPV.cpp
 *
 * Purpose: This file defines the mxcaPV class.
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 2004, 2008-2014 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXCA_PV_ALWAYS_POST_EVENT	FALSE

#define MXCA_PV_DEBUG			FALSE

#define MXCA_PV_DEBUG_READ		FALSE

#define MXCA_PV_DEBUG_WRITE		FALSE

#define MXCA_PV_DEBUG_CALLBACK		FALSE

#define MXCA_PV_DEBUG_GDD_DUMP		FALSE

#define MXCA_PV_DEBUG_SHOW_GDD_VALUE	FALSE

#define MXCA_PV_DEBUG_ENUM		FALSE

#define MXCA_PV_DEBUG_INTEREST		FALSE

#define MXCA_PV_DEBUG_CHECK_FOR_EVENT	FALSE

#define MXCA_PV_DEBUG_VALUE_INITIALIZED FALSE

#define MXCA_PV_DEBUG_UNIMPLEMENTED_APPLICATION_TYPES	FALSE

/* FIXME - Currently the unimplemented application types include
 * the graphic, control, and alarm limits, as well as the units.
 */

#include <stdio.h>
#include <typeinfo>

#include "mxcaServer.h"

#include "dbMapper.h"
#include "gddApps.h"

extern "C" {
#include "mx_time.h"
#include "mx_hrt.h"
}

static aitEnum
mxca_get_primitive_type_from_mx_field_type( long datatype )
{
	aitEnum primitive_type;

	switch( datatype ) {
	case MXFT_STRING:
		primitive_type = aitEnumString;
		break;
	case MXFT_CHAR:
		primitive_type = aitEnumInt8;
		break;
	case MXFT_UCHAR:
		primitive_type = aitEnumUint8;
		break;
	case MXFT_SHORT:
		primitive_type = aitEnumInt16;
		break;
	case MXFT_USHORT:
		primitive_type = aitEnumUint16;
		break;
	case MXFT_BOOL:
	case MXFT_LONG:
		primitive_type = aitEnumInt32;
		break;
	case MXFT_ULONG:
	case MXFT_HEX:
		primitive_type = aitEnumUint32;
		break;
	case MXFT_FLOAT:
		primitive_type = aitEnumFloat32;
		break;
	case MXFT_DOUBLE:
		primitive_type = aitEnumFloat64;
		break;
	default:
		primitive_type = aitEnumInvalid;
		break;
	}
	return primitive_type;
}

/*-------------*/

static void
mxca_show_gdd_value( FILE *file, gdd *gdd_object )
{
	static const char fname[] = "mxca_show_gdd_value()";

	unsigned long i, num_elements, max_items_to_print, items_to_print;
	void *data_pointer;

	max_items_to_print = 20;

	if ( file == (FILE *) NULL ) {
		MX_DEBUG(-2,("%s: FILE pointer passed was NULL.", fname ));
		return;
	}

	if ( gdd_object == (gdd *) NULL ) {
		fprintf( file, "(null object)" );
		return;
	}

#if MXCA_PV_DEBUG_SHOW_GDD_VALUE
	MX_DEBUG(-2,("%s: dumping object %p", fname, gdd_object));
	gdd_object->dump();
#endif

	/* Handle containers first. */

	if ( gdd_object->isContainer() ) {
		fprintf( file, "(CONTAINER)" );
		return;
	}

	/* Handle Scalar and Atomic objects. */

	num_elements = gdd_object->getDataSizeElements();

#if MXCA_PV_DEBUG_SHOW_GDD_VALUE
	MX_DEBUG(-2,("%s: num_elements = %lu", fname, num_elements));
#endif

	if ( num_elements > max_items_to_print ) {
		items_to_print = max_items_to_print;
	} else {
		items_to_print = num_elements;
	}

	data_pointer = gdd_object->dataPointer();

	if ( data_pointer == NULL ) {
		fprintf( file, "(null data)\n" );
		return;
	}

	switch( gdd_object->primitiveType() ) {
	case aitEnumInvalid:
		fprintf( file, "(aitEnumInvalid)" );
		break;

	case aitEnumInt8:
		if ( gdd_object->isScalar() ) {
			union {
				void *pointer;
				aitInt8 int8;
			} ui8;

			ui8.pointer = data_pointer;

			fprintf( file, "%u", (unsigned int) ui8.int8 );
		} else {
			aitInt8 *i8 = (aitInt8 *) data_pointer;

			for ( i = 0; i < items_to_print; i++ ) {
				if ( i > 0 ) {
					fprintf( file, " " );
				}
				fprintf( file, "%d", (int) i8[i] );
			}
		}
		break;

	case aitEnumInt16:
		if ( gdd_object->isScalar() ) {
			union {
				void *pointer;
				aitInt16 int16;
			} ui16;

			ui16.pointer = data_pointer;

			fprintf( file, "%u", (unsigned int) ui16.int16 );
		} else {
			aitInt16 *i16 = (aitInt16 *) data_pointer;

			for ( i = 0; i < items_to_print; i++ ) {
				if ( i > 0 ) {
					fprintf( file, " " );
				}
				fprintf( file, "%d", (int) i16[i] );
			}
		}
		break;

	case aitEnumInt32:
		if ( gdd_object->isScalar() ) {
			union {
				void *pointer;
				aitInt32 int32;
			} ui32;

			ui32.pointer = data_pointer;

			fprintf( file, "%u", (unsigned int) ui32.int32 );
		} else {
			aitInt32 *i32 = (aitInt32 *) data_pointer;

			for ( i = 0; i < items_to_print; i++ ) {
				if ( i > 0 ) {
					fprintf( file, " " );
				}
				fprintf( file, "%d", (int) i32[i] );
			}
		}
		break;

	case aitEnumUint8:
		if ( gdd_object->isScalar() ) {
			union {
				void *pointer;
				aitUint8 uint8;
			} uu8;

			uu8.pointer = data_pointer;

			fprintf( file, "%u", (unsigned int) uu8.uint8 );
		} else {
			aitUint8 *u8 = (aitUint8 *) data_pointer;

			for ( i = 0; i < items_to_print; i++ ) {
				if ( i > 0 ) {
					fprintf( file, " " );
				}
				fprintf( file, "%u", (unsigned int) u8[i] );
			}
		}
		break;

	case aitEnumUint16:
		if ( gdd_object->isScalar() ) {
			union {
				void *pointer;
				aitUint16 uint16;
			} uu16;

			uu16.pointer = data_pointer;

			fprintf( file, "%u", (unsigned int) uu16.uint16 );
		} else {
			aitUint16 *u16 = (aitUint16 *) data_pointer;

			for ( i = 0; i < items_to_print; i++ ) {
				if ( i > 0 ) {
					fprintf( file, " " );
				}
				fprintf( file, "%u", (unsigned int) u16[i] );
			}
		}
		break;

	case aitEnumEnum16:
		if ( gdd_object->isScalar() ) {
			union {
				void *pointer;
				aitEnum16 enum16;
			} ue16;

			ue16.pointer = data_pointer;

			fprintf( file, "%u", (unsigned int) ue16.enum16 );
		} else {
			aitEnum16 *e16 = (aitEnum16 *) data_pointer;

			for ( i = 0; i < items_to_print; i++ ) {
				if ( i > 0 ) {
					fprintf( file, " " );
				}
				fprintf( file, "%u", (unsigned int) e16[i] );
			}
		}
		break;

	case aitEnumUint32:
		if ( gdd_object->isScalar() ) {
			union {
				void *pointer;
				aitUint32 uint32;
			} uu32;

			uu32.pointer = data_pointer;

			fprintf( file, "%u", uu32.uint32 );
		} else {
			aitUint32 *u32 = (aitUint32 *) data_pointer;

			for ( i = 0; i < items_to_print; i++ ) {
				if ( i > 0 ) {
					fprintf( file, " " );
				}
				fprintf( file, "%u", (unsigned int) u32[i] );
			}
		}
		break;

	case aitEnumFloat32:
		if ( gdd_object->isScalar() ) {
			union {
				void *pointer;
				aitFloat32 float32;
			} uf32;

			uf32.pointer = data_pointer;

			fprintf( file, "%g", uf32.float32 );
		} else {
			aitFloat32 *f32 = (aitFloat32 *) data_pointer;

			for ( i = 0; i < items_to_print; i++ ) {
				if ( i > 0 ) {
					fprintf( file, " " );
				}
				fprintf( file, "%g", f32[i] );
			}
		}
		break;

	case aitEnumFloat64:
		if ( gdd_object->isScalar() ) {

			/* The size of a pointer may be smaller than 64-bits,
			 * so we handle 64-bit floating point specially.
			 */

			aitFloat64 float64_data;

			gdd_object->get( aitEnumFloat64, &float64_data );

			fprintf( file, "%g", float64_data );
		} else {
			aitFloat64 *f64 = (aitFloat64 *) data_pointer;

			for ( i = 0; i < items_to_print; i++ ) {
				if ( i > 0 ) {
					fprintf( file, " " );
				}
				fprintf( file, "%g", f64[i] );
			}
		}
		break;

	case aitEnumFixedString:
		if ( gdd_object->isScalar() ) {
			fprintf( file, "%s", (char *) data_pointer );
		} else {
			aitFixedString *fs = (aitFixedString *) data_pointer;

			for ( i = 0; i < items_to_print; i++ ) {
				if ( i > 0 ) {
					fprintf( file, " " );
				}
				fprintf( file, "'%s'", (char *) fs );
			}
		}

	case aitEnumString:
		if ( gdd_object->isScalar() ) {
			fprintf( file, "%s", (char *) data_pointer );
		} else {
			aitString *str = (aitString *) data_pointer;

			for ( i = 0; i < items_to_print; i++ ) {
				if ( i > 0 ) {
					fprintf( file, " " );
				}
				fprintf( file, "'%s'", (char *) str );
			}
		}

	case aitEnumContainer:
		break;
	}

	if ( num_elements > max_items_to_print ) {
		fprintf( file, " ..." );
	}

	return;
}

/*-------------*/

mxcaPV::mxcaPV( mxcaServer *epics_server,
		mxcaRecord *epics_record,
		const char *epics_field_name ) : casPV()
{
	this->create( epics_server, epics_record, epics_field_name );
}

/*- - -*/

mxcaPV::mxcaPV( mxcaRecord *epics_record,
		const char *epics_field_name ) : casPV()
{
	this->create( NULL, epics_record, epics_field_name );
}

/*- - -*/

mxcaPV::mxcaPV( mxcaRecord *epics_record,
		const char *epics_field_name,
		const char *mx_field_name ) : casPV()
{
	this->create( NULL, epics_record, epics_field_name );

	this->configure_mx_field_pv( mx_field_name );
}

/*- - -*/

mxcaPV::mxcaPV( mxcaServer *epics_server,
		const char *epics_field_name ) : casPV()
{
	this->create( epics_server, NULL, epics_field_name );
}

/*-------------*/

void
mxcaPV::create( mxcaServer *epics_server,
		mxcaRecord *epics_record,
		const char *epics_field_name )
{
	static const char fname[] = "mxcaPV::mxcaPV()";

	mxcaPV *PVList;
	mxcaPV *previousPV, *nextPV;

	if ( epics_field_name == NULL ) {
		(void) mx_error( MXE_NULL_ARGUMENT, fname,
		"The epics_field_name pointer passed was NULL." );

		return;		/* FIXME - replace with throw. */
	}

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for epics_server = %p, epics_record = %p, "
		"epics_field_name = '%s'",
			fname, epics_server, epics_record, epics_field_name));
#endif
	if ( epics_server == (mxcaServer *) NULL ) {
		if ( epics_record == (mxcaRecord *) NULL ) {
			(void) mx_error( MXE_NULL_ARGUMENT, fname,
			"Both the mxcaServer and the mxcaRecord pointers "
			"are NULL for EPICS field name '%s'.",
				epics_field_name );

			return;		/* FIXME - replace with throw. */
		} else {
			this->epicsServer = epics_record->epicsServer;
		}
	} else {
		this->epicsServer = epics_server;
	}

	this->epicsRecord = epics_record;

	if ( this->epicsRecord == (mxcaRecord *) NULL ) {
		this->mx_record = NULL;
	} else {
		this->mx_record = this->epicsRecord->mx_record;
	}

	/* If the epics_server argument passed to this method was not NULL,
	 * then we take the epics_field_name to be the entire PV name.
	 *
	 * If the epics_server argument passed to this method _WAS_ NULL,
	 * then the epics_field_name argument is taken to be the field name
	 * of a PV belonging to the epics_record object.  We construct that
	 * field name by concatenating the epics_field_name argument to the
	 * record name of the EPICS record.
	 */

	if ( epics_server != (mxcaServer *) NULL ) {
		strlcpy( this->PVname, epics_field_name, sizeof(this->PVname) );
	} else {
		snprintf( this->PVname, sizeof(this->PVname),
			"%s%s",
			epics_record->epicsRecordName,
			epics_field_name );
	}

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s: EPICS PV name = '%s'\n", fname, this->PVname ));
#endif

	this->set_pv_value_initialized( FALSE );

	/* Add the new PV to the PV list. */

	PVList = this->epicsServer->PVList;

	if ( PVList == (mxcaPV *) NULL ) {

		/* If there is no PVList yet, point nextPV and previousPV
		 * at the new object.
		 */

		this->nextPV = this;
		this->previousPV = this;
	} else {

		nextPV = PVList;

		previousPV = nextPV->previousPV;

		this->nextPV = nextPV;
		this->previousPV = previousPV;

		nextPV->previousPV = this;
		previousPV->nextPV = this;
	}

	this->mx_record_field = NULL;
	this->epics_value = NULL;
	this->mx_value = NULL;
	this->epics_best_external_type = aitEnumInvalid;
	this->num_dimensions = 0;
	this->dimension_array = NULL;
	this->must_transform_data = FALSE;
	this->value_changed_post_in_progress = FALSE;
	this->value_is_monitored = FALSE;
	this->value_changed = FALSE;
	this->value_changed_manual_override = FALSE;
	this->saved_value_changed_test_function = NULL;

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s complete for EPICS PV '%s'\n", fname, this->PVname ));
#endif

	return;
}

mxcaPV::~mxcaPV()
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::~mxcaPV()";

	MX_DEBUG(-2,("%s destructor invoked for PV '%s'", fname, this->PVname));
#endif

	return;
}

/* By default, the Channel Access Server library will destroy any PVs that
 * do not have any channels attached to them.  Given that we precreate all
 * of our PVs, we do not want this to happen.  The standard way of doing
 * this is to provide a 'destroy' method that does nothing.
 */

void
mxcaPV::destroy()
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::destroy()";

	MX_DEBUG(-2,("%s: attempt to destroy PV '%s' intercepted.",
		fname, this->PVname));
#endif

	return;
}

const char *
mxcaPV::getName() const
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::getName()";

	MX_DEBUG(-2,("%s: this->PVname = '%s'", fname, this->PVname));
#endif

	return this->PVname;
}

unsigned
mxcaPV::maxSimultAsyncOps( void ) const
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::maxSimultAsyncOps()";

	MX_DEBUG(-2,("%s invoked for '%s'", fname, this->PVname));
#endif

	return 5000;
}

casChannel *
mxcaPV::createChannel( const casCtx &ctx,
			const char * const user,
			const char * const host )
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::createChannel()";

	MX_DEBUG(-2,("%s invoked for PV '%s', user '%s', host '%s'",
		fname, this->PVname, user, host));
#endif

	mxcaChannel *channel = new mxcaChannel( ctx, user, host, this );

	return channel;
}

void
mxcaPV::genCopy( void *value )
{
	this->epics_value->genCopy( this->epics_best_external_type, value );
}

aitEnum
mxcaPV::bestExternalType() const
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::bestExternalType()";

	MX_DEBUG(-2,("%s: '%s', this->epics_best_external_type = %d",
		fname, this->PVname, this->epics_best_external_type));
#endif

	return this->epics_best_external_type;
}

unsigned
mxcaPV::maxDimension( void ) const
{
	if ( this->dimension_array == NULL ) {
		return 0;
	} else {
		return this->num_dimensions;
	}
}

aitIndex
mxcaPV::maxBound( unsigned dimension ) const
{
	unsigned dimension_size;

	if ( this->dimension_array == NULL ) {
		return 0;
	} else {
		if ( ((int) dimension) > this->num_dimensions ) {
			return 0;
		} else
		if ( dimension == 0 ) {
			dimension_size = this->dimension_array[0];

			return dimension_size;
		} else {
			return 0;
		}
	}
}

/*------------------------------------------------------------------------*/

static mx_status_type
mxp_local_field_callback_function( MX_CALLBACK *callback_struct,
				void *callback_argument )
{
	static const char fname[] = "mxp_local_field_callback_function()";

	mxcaPV *pv;

	if ( callback_argument == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The callback_argument pointer passed was NULL." );
	}

	pv = (mxcaPV *) callback_argument;

#if MXCA_PV_DEBUG_CALLBACK
	MX_DEBUG(-2,("%s invoked for PV '%s'.", fname, pv->getName() ));
#endif

	MX_RECORD_FIELD *mx_field = pv->get_mx_field();

	if ( mx_field == (MX_RECORD_FIELD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RECORD_FIELD pointer for EPICS PV '%s' is NULL.",
			pv->getName() );
	}

#if MXCA_PV_DEBUG_CALLBACK
	MX_DEBUG(-2,("%s: MX field = '%s.%s'",
		fname, mx_field->record->name, mx_field->name));
#endif

	/* If we get here, then the underlying MX field's value has changed,
	 * so we post an event to EPICS clients with subscriptions to this PV
	 * to tell them about this.
	 */

	mx_bool_type post_status = pv->post_value_changed();

#if MXCA_PV_DEBUG_CALLBACK
	MX_DEBUG(-2,("%s: '%s' post_value_changed() status = %d",
		fname, pv->getName(), (int) post_status));
#else
	post_status = post_status;
#endif

	return MX_SUCCESSFUL_RESULT;
}

/*------------------------------------------------------------------------*/

extern "C" {

static mx_status_type
mxp_local_field_callback_shim( MX_CALLBACK *callback_struct,
				void *callback_argument )
{
	mx_status_type mx_status;

#if MXCA_PV_DEBUG_CALLBACK
	MX_DEBUG(-2,( "mxp_local_field_callback_shim(): "
			"callback_struct = %p, callback_argument = %p",
			callback_struct, callback_argument ));
#endif

	/* Pass this on to a C++ function that can handle the mxcaPV
	 * object pointed to by 'callback_argument'.
	 */

	mx_status = mxp_local_field_callback_function( callback_struct,
							callback_argument );

	return mx_status;
}

}  /* End of extern "C" */

/*------------------------------------------------------------------------*/

caStatus
mxcaPV::interestRegister( void )
{
#if MXCA_PV_DEBUG_INTEREST
	static const char fname[] = "mxcaPV::interestRegister()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif
	MX_RECORD_FIELD *mx_field;
	MX_CALLBACK *callback_struct;
	mx_status_type mx_status;

	/* Our first EPICS client has just started monitoring us, so save
	 * that fact.
	 */

	this->value_is_monitored = TRUE;

	/* If the value of this PV has never been initialized from MX,
	 * then read it now.  pv_value_initialized will be set
	 * to TRUE along the way during the call to this->read().
	 */

	if ( this->get_pv_value_initialized() == FALSE ) {
		this->value_changed_manual_override = TRUE;

		this->check_for_event();

		this->set_pv_value_initialized( TRUE );
	}

	/* Does this PV have an associated MX field? */

	mx_field = this->get_mx_field();

#if MXCA_PV_DEBUG_INTEREST
	MX_DEBUG(-2,("%s: mx_field = %p", fname, mx_field));
#endif

	if ( mx_field == (MX_RECORD_FIELD *) NULL ) {

		/* If it does not, then we are done here. */

		return S_casApp_success;
	}

	/* If it _does_ have one, then add a local callback so that
	 * we will be able to post events to subscribers for this
	 * EPICS PV.
	 *
	 * Note that the callback function _must_ have C linkage.
	 * However, the callback function must also arrange to invoke
	 * the C++ post_value_changed() method.  This means that we
	 * must supply the PVs object pointer as an argument to the
	 * mx_local_field_add_new_callback(), but not make use of
	 * that object pointer until we are back in C++ land.
	 */

#if MXCA_PV_DEBUG_INTEREST
	MX_DEBUG(-2,
	("%s: Invoking mx_local_field_add_callback() for field '%s.%s'",
		fname, mx_field->record->name, mx_field->name));
#endif

	mx_status = mx_local_field_add_new_callback( mx_field, MXCBT_POLL,
						mxp_local_field_callback_shim,
						this,
						&callback_struct );

#if MXCA_PV_DEBUG_INTEREST
	MX_DEBUG(-2,("%s: PV = %p, callback_struct = %p",
		fname, this, callback_struct));
#endif

	if ( mx_status.code != MXE_SUCCESS )
		return S_casApp_undefined;

	return S_casApp_success;
}

void
mxcaPV::interestDelete( void )
{
#if MXCA_PV_DEBUG_INTEREST
	static const char fname[] = "mxcaPV::interestDelete()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname ));
#endif

	/* It appears that if nobody is monitoring this PV and somebody
	 * reads() or writes() the PV, then the interestDelete() method
	 * is invoked even though a matching call to interestRegister()
	 * was not made.
	 *
	 * The symptom of this phenomenon is that this method is called
	 * with this->value_is_monitored set to FALSE.  If this happens,
	 * then just return to our caller without doing anything.
	 */

	if ( this->value_is_monitored == FALSE ) {
		return;
	}

	/* We were being monitored, but a client has stopped monitoring
	 * us.  To reflect this, we set 'value_is_monitored' to FALSE.
	 */

	this->value_is_monitored = FALSE;

	return;
}

void
mxcaPV::check_for_event( void )
{
	static const char fname[] = "mxcaPV::check_for_event()";

	mx_bool_type value_changed;

#if MXCA_PV_DEBUG_CHECK_FOR_EVENT
	MX_DEBUG(-2,("%s Invoked for PV '%s' vvvvvvvvvvvvvvvvvvvv",
		fname, this->PVname));
#endif

	if ( this->epics_value == (gdd *) NULL ) {

#if MXCA_PV_DEBUG_CHECK_FOR_EVENT
		MX_DEBUG(-2,("%s: skipping PV '%s' since it has no value.",
			fname, this->PVname));
#endif
		return;
	}

	/* Poll the hardware for its current value. */

	caStatus epics_status = this->read();

	if ( epics_status != S_casApp_success) {
		(void) mx_error( MXE_FUNCTION_FAILED, fname,
		"The attempt to read from PV '%s' failed.  "
		"EPICS status = %d, EPICS error = '%s'.",
			this->PVname, epics_status, ca_message( epics_status ));
	}

	this->set_status_severity_and_timestamp( 0, 0 );

	/* If the value has changed, then post an event. */

#if MXCA_PV_ALWAYS_POST_EVENT
	value_changed = TRUE;
#else
	value_changed = this->value_changed;
#endif
	if ( this->value_changed_manual_override ) {
		value_changed = TRUE;

		this->value_changed_manual_override = FALSE;

#if MXCA_PV_DEBUG_CHECK_FOR_EVENT
		MX_DEBUG(-2,("%s: Manual override requested for PV '%s'",
			fname, this->PVname));
#endif
	}

	if ( value_changed ) {
		this->post_value_changed();
	}

	this->value_changed = FALSE;

#if MXCA_PV_DEBUG_CHECK_FOR_EVENT
	MX_DEBUG(-2,("%s Complete for PV '%s' ^^^^^^^^^^^^^^^^^^^^^",
		fname, this->PVname));
#endif

	return;
}

mx_bool_type
mxcaPV::post_value_changed( void )
{
#if MXCA_PV_DEBUG_CHECK_FOR_EVENT
	static const char fname[] = "mxcaPV::post_value_changed()";
#endif

	mxcaServer *epics_server;

	if ( this->value_changed_post_in_progress ) {

#if MXCA_PV_DEBUG_CHECK_FOR_EVENT
		MX_DEBUG(-2,("%s skipped for PV '%s'", fname, this->PVname));
#endif
		return FALSE;
	}

#if MXCA_PV_DEBUG_CHECK_FOR_EVENT
	MX_DEBUG(-2,("%s called for PV '%s'", fname, this->PVname));
#endif

	epics_server = this->epicsServer;

	/* If requested, display EPICS post value changed debug message. */

	if ( epics_server->epics_network_debug_flags ) {
		fprintf( stderr, "EPICS POST_EVENT('%s') = ", this->PVname );
		mxca_show_gdd_value( stderr, this->epics_value );
		fprintf( stderr, "\n" );
	}

	/* Post the event to remote EPICS clients via subscriptions. */

	this->postEvent( epics_server->event_mask, *(this->epics_value) );

	return TRUE;
}

void
mxcaPV::set_status_severity_and_timestamp( aitInt16 status, aitInt16 severity )
{
#if MXCA_PV_DEBUG
	static const char fname[] =
		"mxcaPV::set_status_severity_and_timestamp()";
#endif

	struct timespec current_timespec;

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	/* If the PV has no 'epics_value' GDD structure, then there is
	 * nothing we can do here.
	 */

	if ( this->epics_value == (gdd *) NULL ) {

#if MXCA_PV_DEBUG
		MX_DEBUG(-2,
		("%s: PV '%s' has no 'epics_value' structure.  Skipping...",
		 	fname, this->PVname));
#endif

		return;
	}

	/* Set EPICS status and severity. */

	this->epics_value->setStatSevr( status, severity );

	/* Set the EPICS timestamp. */

	current_timespec = mx_current_os_time();

	this->epics_value->setTimeStamp( &current_timespec );

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return;
}

caStatus
mxcaPV::read( const casCtx &ctx, gdd &gdd_object )
{
#if 1 || MXCA_PV_DEBUG_READ || MXCA_PV_DEBUG_UNIMPLEMENTED_APPLICATION_TYPES
	static const char fname[] = "mxcaPV::read( const casCtx &, gdd & )";
#endif

	caStatus read_status;

#if MXCA_PV_DEBUG_READ
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	if ( gdd_object.isContainer() ) {

#if MXCA_PV_DEBUG_READ
		MX_DEBUG(-2,("%s: isContainer() == TRUE", fname));
#endif
		gdd *current_gdd;
		aitUint32 i;

		aitUint32 num_elements = gdd_object.getDataSizeElements();

		gddContainer &container = (gddContainer &) gdd_object;

		gddCursor cursor = container.getCursor();

		cursor.first();

		for ( i = 0; i < num_elements; i++ ) {
			current_gdd = cursor.current();

#if MXCA_PV_DEBUG_READ
			MX_DEBUG(-2,("%s: examining gdd %lu.",
				fname, (unsigned long) i));
#endif

			/* Recursively invoke the read method on
			 * the current gdd.
			 */

			read_status = this->read( ctx, *current_gdd );

			cursor.next();
		}
		
		return S_casApp_success;
	}

	/* Decide what to do based on the GDD's application type. */

	unsigned application_type = gdd_object.applicationType();

#if MXCA_PV_DEBUG_READ
	MX_DEBUG(-2,("%s: application_type = %u", fname, application_type));
#endif
	switch( application_type ) {
	case gddAppType_value:
		/* 'value' is the most commonly used application type. */

		read_status = this->read( gdd_object );
		break;
	case gddAppType_precision:
		read_status = this->read_precision( gdd_object );
		break;
	case gddAppType_enums:
		read_status = this->read_enums( gdd_object );
		break;
	case gddAppType_class:
		(void) mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"Support for 'PV Info' is not yet implemented for PV '%s'.",
			this->PVname );

		read_status = S_casApp_noSupport;
		break;
	default:

#if MXCA_PV_DEBUG_UNIMPLEMENTED_APPLICATION_TYPES
		(void) mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"Unimplemented application type %u requested for PV '%s'.",
			application_type, this->PVname );
#endif

		read_status = S_casApp_noSupport;
		break;
	}

#if MXCA_PV_DEBUG_READ
	MX_DEBUG(-2,("%s complete, read_status = %d", fname, read_status));
#endif
	return read_status;
}

caStatus
mxcaPV::read( gdd &gdd_object )
{
#if MXCA_PV_DEBUG_READ
	static const char fname[] = "mxcaPV::read( gdd & )";
#endif

	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();

#if MXCA_PV_DEBUG_READ
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	caStatus epics_status = this->read();

	if ( epics_status == S_casApp_success ) {

		/* Copy to the object to send to the client. */

		table.smartCopy( &gdd_object, this->epics_value );

#if MXCA_PV_DEBUG_GDD_DUMP
		MX_DEBUG(-2,("%s: dumping object %p", fname, &gdd_object ));
		gdd_object.dump();
#endif
	}

#if MXCA_PV_DEBUG_READ
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return epics_status;
}

caStatus
mxcaPV::read( void )
{
#if MXCA_PV_DEBUG_READ
	static const char fname[] = "mxcaPV::read( void )";
#endif

	int use_mx_record_processing;
	mx_status_type mx_status;

#if MXCA_PV_DEBUG_READ
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	/* Do we need to do MX record processing? */

	if ( ( this->mx_record == (MX_RECORD *) NULL )
	  || ( this->mx_record_field == (MX_RECORD_FIELD *) NULL ) )
	{
		use_mx_record_processing = FALSE;
	} else {
		use_mx_record_processing = TRUE;
	}

#if MXCA_PV_DEBUG_READ
	MX_DEBUG(-2,("%s: use_mx_record_processing = %d",
		fname, use_mx_record_processing));
#endif

	if ( use_mx_record_processing == FALSE ) {
		if ( this->get_pv_value_initialized() == FALSE ) {
			mx_warning(
		    "PV '%s' has no MX record field and needs initialization.",
				this->PVname );

			this->set_pv_value_initialized( TRUE );
		}
	} else {
		if ( this->get_pv_value_initialized() == FALSE ) {
			this->set_pv_value_initialized( TRUE );
		}

#if MXCA_PV_DEBUG_READ
		MX_DEBUG(-2,("%s: Processing record field '%s.%s'", fname, 
			this->mx_record->name, this->mx_record_field->name));
#endif

		/* Read the values from the hardware. */

		mx_status = mx_process_record_field( this->mx_record,
							this->mx_record_field,
							MX_PROCESS_GET,
							&(this->value_changed));
#if 0
		if ( 1 ) {
			fprintf( stderr, "%s---> Field '%s.%s' = ",
				fname, this->mx_record->name,
				this->mx_record_field->name );

			mx_print_field( stderr,
					this->mx_record,
					this->mx_record_field,
					TRUE );

			fprintf( stderr, "\n" );
		}
#endif

		/* If the MX value must be transformed before sending it
		 * to EPICS, perform the transformation here.
		 */

#if MXCA_PV_DEBUG_READ
		MX_DEBUG(-2,("%s: this->must_transform_data = %d",
			fname, this->must_transform_data));
#endif

		if ( this->must_transform_data ) {
			mx_status = this->transform_mx_data( MX_PROCESS_GET );
		}

#if MXCA_PV_DEBUG_READ
		MX_DEBUG(-2,("%s: genCopy() to this->mx_value = %p",
			fname, this->mx_value));
#endif
		if ( this->epics_value->primitiveType() == aitEnumString ) {

			aitString &string_object =
				(aitString &) this->epics_value->getData();

			string_object.copy( (const char *) this->mx_value );
		} else {
			this->epics_value->genCopy(
				this->epics_best_external_type,
				this->mx_value );
		}
	}

	/* Make sure the PV status, etc. is up to date. */

	this->set_status_severity_and_timestamp( 0, 0 );

	/* If requested, display EPICS network read debug message. */

	if ( this->epicsServer->epics_network_debug_flags ) {
		fprintf( stderr, "EPICS READ('%s') = ", this->PVname );
		mxca_show_gdd_value( stderr, this->epics_value );
		fprintf( stderr, "\n" );
	}

#if MXCA_PV_DEBUG_READ
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return S_casApp_success;
}

caStatus
mxcaPV::write( const casCtx &ctx, const gdd &gdd_object )
{
#if MXCA_PV_DEBUG_WRITE
	static const char fname[] =
		"mxcaPV::write( const casCtx &, const gdd & )";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	caStatus write_status = this->write( gdd_object );

#if MXCA_PV_DEBUG_WRITE
	MX_DEBUG(-2,("%s: write_status = %d", fname, write_status));
#endif

	return write_status;
}

caStatus
mxcaPV::write( const gdd &gdd_object )
{
#if MXCA_PV_DEBUG_WRITE
	static const char fname[] = "mxcaPV::write( const gdd & )";
#endif

	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();
	mx_status_type mx_status;

#if MXCA_PV_DEBUG_WRITE
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

#if MXCA_PV_DEBUG_GDD_DUMP
	MX_DEBUG(-2,("%s: dumping object %p", fname, &gdd_object ));
	gdd_object.dump();
#endif

	/* Copy from the object sent by the client. */

	table.smartCopy( this->epics_value, &gdd_object );

#if MXCA_PV_DEBUG_WRITE
	MX_DEBUG(-2,("%s: this->mx_value = %p", fname, this->mx_value));
#endif

	if ( this->mx_value != NULL ) {

		/* Write the values to the hardware. */

#if MXCA_PV_DEBUG_WRITE
		MX_DEBUG(-2,("%s: Copying EPICS data to the MX field.", fname));
#endif

		if ( this->epics_value->primitiveType() == aitEnumString ) {

			long mx_max_string_length =
				this->mx_record_field->dimension[0];

			aitString &string_object =
				(aitString &) this->epics_value->getData();

			const char *epics_string = string_object.string();

			strlcpy( (char *) this->mx_value, epics_string,
						mx_max_string_length );
		} else {
			this->epics_value->get( this->mx_value,
					this->epics_best_external_type );
		}
#if 0
		if ( 1 ) {
			fprintf( stderr, "%s: ---> Field '%s.%s' = ",
				fname, this->mx_record->name,
				this->mx_record_field->name );

			mx_print_field( stderr,
					this->mx_record,
					this->mx_record_field,
					TRUE );

			fprintf( stderr, "\n" );
		}
#endif

		/* If the EPICS value must be transformed before sending it
		 * to MX, perform the transformation here.
		 */

#if MXCA_PV_DEBUG_WRITE
		MX_DEBUG(-2,("%s: this->must_transform_data = %d",
				fname, this->must_transform_data));
#endif

		if ( this->must_transform_data ) {
			mx_status = this->transform_mx_data( MX_PROCESS_PUT );
		}
	}

	caStatus write_status = this->write();

#if MXCA_PV_DEBUG_WRITE
	MX_DEBUG(-2,("%s: write_status = %d", fname, write_status));
#endif

	return write_status;
}

caStatus
mxcaPV::write( void )
{
#if MXCA_PV_DEBUG_WRITE
	static const char fname[] = "mxcaPV::write( void )";
#endif
	mx_status_type mx_status;

#if MXCA_PV_DEBUG_WRITE
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	if ( this->epicsServer->epics_network_debug_flags ) {
		fprintf( stderr, "EPICS WRITE('%s', ", this->PVname );
			mxca_show_gdd_value( stderr, this->epics_value );
			fprintf( stderr, ")\n" );
	}

	if ( ( this->mx_record == (MX_RECORD *) NULL )
	  || ( this->mx_record_field == (MX_RECORD_FIELD *) NULL ) )
	{
		/* If the PV is not attached to an MX record field,
		 * then all we do is post the change to EPICS clients.
		 */

#if MXCA_PV_DEBUG_WRITE
		MX_DEBUG(-2,("%s: Only posting value changed.", fname));
#endif
		this->post_value_changed();
	} else
	if ( this->mx_value != NULL ) {

		if ( this->get_pv_value_initialized() == FALSE ) {
			this->set_pv_value_initialized( TRUE );
		}

		/* Send the value to the MX driver. */

#if MXCA_PV_DEBUG_WRITE
		MX_DEBUG(-2,("%s: Processing record field '%s.%s'", fname,
			this->mx_record->name, this->mx_record_field->name));
#endif

		mx_status = mx_process_record_field( this->mx_record,
						this->mx_record_field,
						MX_PROCESS_PUT,
						&(this->value_changed) );
#if MXCA_PV_DEBUG_WRITE
		MX_DEBUG(-2,("%s: this->value_changed = %d",
			fname, this->value_changed));
#endif
		if ( this->value_changed ) {

#if 0
#if MXCA_PV_DEBUG_WRITE
			MX_DEBUG(-2,("%s: invoking this->genCopy()", fname));
#endif
			this->genCopy( this->mx_value );
#endif

#if MXCA_PV_DEBUG_WRITE
			MX_DEBUG(-2,("%s: Posting value changed for '%s.%s'",
				fname, this->mx_record->name,
				this->mx_record_field->name));
#endif
			this->post_value_changed();
		}
	}

#if MXCA_PV_DEBUG_WRITE
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return S_casApp_success;
}

mx_status_type
mxcaPV::transform_mx_data( int direction )
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::transform_mx_data()";
#endif

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	/* The default transformation is to make no changes to the data. */

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return MX_SUCCESSFUL_RESULT;
}

caStatus
mxcaPV::read_precision( gdd &gdd_object )
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::read_precision()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	/* FIXME: For now we set all precision values to 4. */

	gdd_object.put( 4 );

	return S_casApp_success;
}

caStatus
mxcaPV::read_enums( gdd &gdd_object )
{
	static const char fname[] = "mxcaPV::read_enums()";

	MX_DEBUG(-2,
	("ERROR!!!: %s invoked for PV '%s'.  This should not happen.",
		fname, this->PVname));

	return S_casApp_noSupport;
}

mx_status_type
mxcaPV::configure_internal_pv( aitEnum primitive_type,
				unsigned dimension,
				aitUint32 *size_array,
				const void *initial_value )
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::configure_internal_pv()";
#endif

	int application_type;
	char *string_ptr;

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for EPICS PV '%s'", fname, this->PVname));
#endif

	application_type = gddAppType_value;

	this->epics_best_external_type = primitive_type;

	if ( dimension == 0 ) {

#if MXCA_PV_DEBUG
		MX_DEBUG(-2,("%s: PV '%s', creating gddScalar( %d, %d )",
			fname, this->PVname, application_type, primitive_type));
#endif

		this->epics_value = new gddScalar( application_type,
							primitive_type );
	} else {

#if MXCA_PV_DEBUG
		MX_DEBUG(-2,
		("%s: PV '%s', creating gddAtomic( %d, %d, %d, %d )",
		 	fname, this->PVname, application_type, primitive_type,
					dimension, size_array[0] ));
#endif

		this->epics_value = new gddAtomic( application_type,
							primitive_type,
							dimension, size_array );
	}

	switch( primitive_type ) {
	case aitEnumString:
		string_ptr = (char *) initial_value;

		this->epics_value->genCopy( primitive_type, &string_ptr );
		break;
	default:
		this->epics_value->genCopy( primitive_type, initial_value );
		break;
	}

#if MXCA_PV_DEBUG_GDD_DUMP
	MX_DEBUG(-2,("%s: dumping object %p", fname, this->epics_value ));
	this->epics_value->dump();
#endif

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s complete for EPICS PV '%s'", fname, this->PVname));
#endif

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxcaPV::configure_mx_field_pv( const char *mx_field_name )
{
	static const char fname[] = "mxcaPV::configure_mx_field_pv()";

	MX_RECORD_FIELD *field, *matching_field;
	MX_RECORD *mx_record;
	long i;
	mx_status_type mx_status;

	if ( mx_field_name == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The mx_field_name pointer passed was NULL." );
	}

	mx_record = this->mx_record;

	if ( mx_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"EPICS PV '%s' does not have an associated MX record.",
			this->PVname );
	}

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for EPICS PV '%s', MX field '%s.%s'",
		fname, this->PVname, mx_record->name, mx_field_name ));
#endif

	matching_field = NULL;

	for ( i = 0; i < mx_record->num_record_fields; i++ ) {

		field = &(mx_record->record_field_array[i]);

		if ( strcmp( mx_field_name, field->name ) == 0 ) {
			matching_field = field;

			break;
		}
	}

	if ( matching_field == NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Error in EPICS PV '%s'.  The corresponding MX record '%s' "
		"does not have a field named '%s'",
			this->PVname, mx_record->name, mx_field_name );
	}

	mx_status = this->configure_mx_field_pv( matching_field );

	return mx_status;
}

mx_status_type
mxcaPV::configure_mx_field_pv( MX_RECORD_FIELD *mx_record_field )
{
	static const char fname[] = "mxcaPV::configure_mx_field_pv()";

	MX_RECORD *mx_record;
	aitEnum primitive_type;
	unsigned num_dimensions;
	unsigned long max_string_length;
	int application_type;
	aitUint32 size_array[1];

	if ( mx_record_field == (MX_RECORD_FIELD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD_FIELD pointer passed was NULL." );
	}

	mx_record = mx_record_field->record;

	if ( mx_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RECORD pointer for MX record field %p '%s' is NULL.",
			mx_record_field, mx_record_field->name );
	}

	this->mx_record_field = mx_record_field;

	this->mx_value = mx_get_field_value_pointer( mx_record_field );

	if ( this->mx_record == (MX_RECORD *) NULL ) {
		this->mx_record = mx_record;
	}

	/* Find out what data type this MX field is and create
	 * a matching GDD structure.
	 */

	num_dimensions = mx_record_field->num_dimensions;

	if ( ( num_dimensions != 0 ) && ( num_dimensions != 1 ) ) {
		return mx_error( MXE_UNSUPPORTED, fname,
	"The MX Channel Access server only supports exporting MX record fields "
	"with 0 or 1 dimensions.  The requested field '%s' for record '%s' "
	"has %u dimensions.", mx_record_field->name, mx_record->name,
				num_dimensions );
	}

	/* Get the EPICS primitive type and dimensions for this field. */

	if ( mx_record_field->datatype == MXFT_STRING ) {
		/* EPICS strings are special. */

		max_string_length = mx_record_field->dimension[0];

		if ( max_string_length > 40 ) {

			/* For long strings (> 40), we use an array of chars
			 * for the string.
			 *
			 * WARNING: Not all EPICS clients know how to deal
			 * with treating a char array as a string.
			 */

			primitive_type = aitEnumInt8;
			num_dimensions = 1;
			size_array[0]  = max_string_length;

			this->num_dimensions = 1;
			this->dimension_array = mx_record_field->dimension;
		} else {
			/* For short strings, we use the traditional
			 * EPICS string type.
			 */

			primitive_type = aitEnumString;
			num_dimensions = 0;
		}
	} else {
		/* Numerical fields. */

		if ( num_dimensions == 0 ) {
			this->num_dimensions = 0;
			this->dimension_array = NULL;
		} else
		if ( num_dimensions == 1 ) {
			this->num_dimensions = 1;
			this->dimension_array = mx_record_field->dimension;

			size_array[0] = mx_record_field->dimension[0];
		}

		primitive_type = mxca_get_primitive_type_from_mx_field_type(
					mx_record_field->datatype );

		if ( primitive_type == aitEnumInvalid ) {
			return mx_error( MXE_UNSUPPORTED, fname,
			"MX record field '%s.%s' may not be exported to "
			"EPICS since it is of unsupported type %ld '%s'.",
				mx_record->name, mx_record_field->name,
				mx_record_field->datatype,
			mx_get_field_type_string( mx_record_field->datatype ) );
		}
	}

	this->epics_best_external_type = primitive_type;

	application_type = gddAppType_value;

	/* Now create the gdd object that EPICS uses for the
	 * value of the field.
	 */

	if ( primitive_type == aitEnumFixedString ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"The aitFixedString data type is not supported by this server."
		"  Please use the aitString data type instead." );
	}

	if ( num_dimensions == 0 ) {

#if MXCA_PV_DEBUG
		MX_DEBUG(-2,("%s: PV '%s', creating gddScalar( %d, %d )",
			fname, this->PVname, application_type, primitive_type));
#endif
		if ( primitive_type != aitEnumString ) {
			/* Non-string GDDs can be created directly. */

			this->epics_value = new gddScalar( application_type,
							primitive_type );
		} else {
			/* String GDDs must be treated specially.
			 *
			 * 1. Create an empty aitString object.
			 * 2. Create an empty gddScalar object.
			 * 3. Insert the aitString into the gddScalar
			 *    object.
			 */

			aitString str = new aitString();

			this->epics_value = new gddScalar( application_type,
							aitEnumString );

			this->epics_value->put( str );
		}
	} else
	if ( num_dimensions == 1 ) {

#if MXCA_PV_DEBUG
		MX_DEBUG(-2,
		("%s: PV '%s', creating gddAtomic( %d, %d, %d, %d )",
			fname, this->PVname, application_type, primitive_type,
			num_dimensions, size_array[0] ));
#endif

		this->epics_value = new gddAtomic( application_type,
						primitive_type,
						num_dimensions, size_array );
	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
			"dimension (%u) is greater than 1.  "
			"We should not be able to get here.", num_dimensions );
	}

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,
("%s: '%s', this->mx_value = %p, epics_data_pointer = %p, primitive_type = %d",
		fname, this->PVname, this->mx_value,
		this->epics_value->dataPointer(), primitive_type));

	MX_DEBUG( 2,("%s completed for EPICS PV '%s', MX field '%s.%s'",
		fname, this->PVname, mx_record->name, mx_record_field->name ));
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_RECORD_FIELD *
mxcaPV::get_mx_field( void )
{
	return this->mx_record_field;
}

void *
mxcaPV::get_mx_value_pointer( void ) const
{
	return this->mx_value;
}

void
mxcaPV::set_mx_value_pointer( void *value_ptr )
{
	this->mx_value = value_ptr;
	return;
}

mxcaPV *
mxcaPV::get_next_pv( void ) const
{
	return this->nextPV;
}

mx_bool_type
mxcaPV::get_value_changed( void )
{
	return this->value_changed;
}


mx_bool_type
mxcaPV::get_pv_value_initialized( void )
{
#if MXCA_PV_DEBUG_VALUE_INITIALIZED
	static const char fname[] = "mxcaPV::get_pv_value_initialized()";

	MX_DEBUG(-2,("%s: PV '%s', pv_value_initialized = %d",
		fname, this->PVname, this->pv_value_initialized));
#endif

	return this->pv_value_initialized;
}

void
mxcaPV::set_pv_value_initialized( mx_bool_type new_value )
{
#if MXCA_PV_DEBUG_VALUE_INITIALIZED
	static const char fname[] = "mxcaPV::set_pv_value_initialized()";

	MX_DEBUG(-2,("%s: PV '%s', pv_value_initialized = %d",
		fname, this->PVname, new_value));
#endif

	this->pv_value_initialized = new_value;
}

/*--------------------------------------------------------------------------*/

mxcaEnumPV::mxcaEnumPV( mxcaRecord *epics_record, const char *epics_field_name )
			: mxcaPV( epics_record, epics_field_name )
{
	return;
}

mxcaEnumPV::~mxcaEnumPV( void )
{
	return;
}

mx_status_type
mxcaEnumPV::configure_enum_pv( aitEnum16 initial_value,
				unsigned num_enum_values, ... )
{
#if MXCA_PV_DEBUG_ENUM
	static const char fname[] = "mxcaEnumPV::configure_enum_pv()";
#endif

	va_list args;
	unsigned i;
	char *ptr;
	mx_status_type mx_status;

#if MXCA_PV_DEBUG_ENUM
	MX_DEBUG(-2,("%s invoked for EPICS PV '%s'", fname, this->PVname));
#endif

	this->enum_value = initial_value;
	this->num_enums = num_enum_values;

	this->mx_value = &(this->enum_value);

	va_start(args, num_enum_values);

	this->enum_array = new aitFixedString[ this->num_enums ];

	for ( i = 0; i < this->num_enums; i++ ) {
		ptr = va_arg( args, char * );

		strlcpy( this->enum_array[i].fixed_string,
				ptr, sizeof( this->enum_array[i] ) );

#if MXCA_PV_DEBUG_ENUM
		MX_DEBUG(-2,("%s: this->enum_array[%u].fixed_string = '%s'",
			fname, i, this->enum_array[i].fixed_string));
#endif
	}

	va_end(args);

	mx_status = this->configure_internal_pv( aitEnumEnum16, 0,
						NULL, &(this->enum_value) );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	this->epics_value->putConvert( initial_value );

#if 0
	MX_DEBUG(-2,("%s: dumping object %p", fname, this->epics_value ));
	this->epics_value->dump();
#endif

#if MXCA_PV_DEBUG_ENUM
	MX_DEBUG(-2,("%s complete for EPICS PV '%s'", fname, this->PVname));
#endif

	return MX_SUCCESSFUL_RESULT;
}

caStatus
mxcaEnumPV::read_enums( gdd &gdd_object )
{
#if MXCA_PV_DEBUG_ENUM
	static const char fname[] = "mxcaPV::read_enums()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	gdd_object.setDimension( 1 );
	gdd_object.setBound( 0, 0, this->num_enums );
	gdd_object.putRef( this->enum_array );

	return S_casApp_success;
}

caStatus
mxcaEnumPV::write( const gdd &gdd_object )
{
#if MXCA_PV_DEBUG_ENUM
	static const char fname[] = "mxcaEnumPV::write()";
#endif

	aitEnum16 gdd_value;
	caStatus epics_status;

#if MXCA_PV_DEBUG_ENUM
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	/* Save the value in the this->epics_value structure. */

	epics_status = mxcaPV::write( gdd_object );

	if ( epics_status != S_casApp_success )
		return epics_status;

	/* Get the PV value from the GDD structure so that we may
	 * save a copy of it in this->enum_value.
	 */

	gdd_object.getConvert( gdd_value );

	this->enum_value = gdd_value;

#if MXCA_PV_DEBUG_ENUM
	MX_DEBUG(-2,("%s: PV '%s', this->enum_value = %d",
		fname, this->PVname, this->enum_value));

	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return S_casApp_success;
}

/*--------------------------------------------------------------------------*/

mxcaDummyPV::mxcaDummyPV( mxcaServer *epics_server,
			const char *epics_pvname ) :
			mxcaPV( epics_server, epics_pvname )
{
	;
}

mxcaDummyPV::mxcaDummyPV( mxcaRecord *epics_record,
			const char *epics_pvname ) :
			mxcaPV( epics_record, epics_pvname )
{
	;
}

caStatus
mxcaDummyPV::read( void )
{
	return S_casApp_success;
}

caStatus
mxcaDummyPV::write( const gdd &gdd_object )
{
	return S_casApp_success;
}

