/*
 * Name:    mxcaRecord.cpp
 *
 * Purpose: This file defines the mxcaRecord class and the classes that 
 *          inherit from it.
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 2004, 2008-2009, 2012 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <iostream>
#include <fstream>

#include "mxcaServer.h"

mxcaRecord::mxcaRecord( mxcaServer *epics_server,
			const char *epics_record_name,
			MX_RECORD *mx_record )
{
	mx_status_type mx_status;

	this->epicsServer = epics_server;

	if ( epics_record_name != NULL ) {
		strlcpy( this->epicsRecordName, epics_record_name,
			sizeof( this->epicsRecordName ) );
	}

	this->mx_record = mx_record;

	if ( mx_record != (MX_RECORD *) NULL ) {
		mx_record->application_ptr = this;

		if ( mx_record->mx_superclass != MXR_LIST_HEAD ) {
			mx_status = this->setupRecordPVs();
		}
	}
}

mx_status_type
mxcaRecord::setupRecordPVs( void )
{
	mxcaPV *pv;

	pv = new mxcaPV( this, ".NAME", "name" );

	pv = new mxcaPV( this, ".DESC", "label" );

	pv = pv;	/* Suppress GCC 4.6 'set but not used' warning. */

	return MX_SUCCESSFUL_RESULT;
}

const char *
mxcaRecord::get_epics_record_name( void )
{
	return this->epicsRecordName;
}

mxcaPV *
mxcaRecord::findPV( const char *field_name )
{
	mxcaPV *pv;
	char pvname[80];

	if ( this->epicsServer == (mxcaServer *) NULL ) {
		return NULL;
	}

	snprintf( pvname, sizeof(pvname), "%s.%s",
		this->epicsRecordName, field_name );

	pv = this->epicsServer->findPV( pvname );

	return pv;
}

